from typing import Dict, Tuple

import pandas as pd
import pytest

from great_expectations.core.metric_function_types import (
    MetricPartialFunctionTypeSuffixes,
    SummarizationMetricNameSuffixes,
)
from great_expectations.execution_engine import (
    PandasExecutionEngine,
    SparkDFExecutionEngine,
    SqlAlchemyExecutionEngine,
)
from great_expectations.self_check.util import (
    build_pandas_engine,
    build_sa_execution_engine,
    build_spark_engine,
)
from great_expectations.validator.computed_metric import MetricValue
from great_expectations.validator.metric_configuration import MetricConfiguration
from tests.expectations.test_util import get_table_columns_metric


@pytest.fixture
def animal_table_df() -> pd.DataFrame:
    """
    Returns: pandas dataframe that contains example data for unexpected_index_column_names metric tests
    """  # noqa: E501 # FIXME CoP
    df = pd.DataFrame(
        {
            "pk_1": [0, 1, 2, 3, 4, 5],
            "pk_2": ["zero", "one", "two", "three", "four", "five"],
            "animals": [
                "cat",
                "fish",
                "dog",
                "giraffe",
                "lion",
                "zebra",
            ],
        }
    )
    return df


@pytest.fixture
def metric_value_kwargs_complete() -> dict:
    """
    Test configuration for metric_value_kwargs. Contains `unexpected_index_column_names` key
    """
    return {
        "value_set": ["cat", "fish", "dog"],
        "parse_strings_as_datetimes": False,
        "result_format": {
            "result_format": "COMPLETE",
            "unexpected_index_column_names": ["pk_1"],
            "partial_unexpected_count": 20,
            "include_unexpected_rows": False,
            "exclude_unexpected_values": False,
        },
    }


def _build_table_columns_and_unexpected(
    engine, metric_value_kwargs
) -> Tuple[MetricConfiguration, MetricConfiguration, dict]:
    """
    Helper method build the dependencies needed for unexpected_indices (ID/PK) related metrics.

    The unexpected_indices related metrics are dependent on :
        - unexpected_condition
        - table_columns

    This method takes in the engine (connected to data), and metric value kwargs dictionary to build the metrics
    and return as a tuple


    Args:
        engine (ExecutionEngine): has connection to data
        metric_value_kwargs (dict): can be

    Returns:

        Tuple with MetricConfigurations corresponding to unexpected_condition and table_columns metric, as well as metrics dict.

    """  # noqa: E501 # FIXME CoP
    metrics: Dict[Tuple[str, str, str], MetricValue] = {}

    # get table_columns_metric
    table_columns_metric: MetricConfiguration
    results: Dict[Tuple[str, str, str], MetricValue]
    table_columns_metric, results = get_table_columns_metric(execution_engine=engine)
    metrics.update(results)

    # unexpected_condition metric
    unexpected_condition_metric: MetricConfiguration = MetricConfiguration(
        metric_name=f"column_values.in_set.{MetricPartialFunctionTypeSuffixes.CONDITION.value}",
        metric_domain_kwargs={"column": "animals"},
        metric_value_kwargs=metric_value_kwargs,
    )
    unexpected_condition_metric.metric_dependencies = {
        "table.columns": table_columns_metric,
    }
    metrics = engine.resolve_metrics(
        metrics_to_resolve=(unexpected_condition_metric,), metrics=metrics
    )

    return table_columns_metric, unexpected_condition_metric, metrics


@pytest.mark.unit
def test_pd_unexpected_index_list_metric_without_id_pk(animal_table_df):
    df: pd.DataFrame = animal_table_df
    # pandas will return default unexpected_index_list without id_pk
    metric_value_kwargs: dict = {
        "value_set": ["cat", "fish", "dog"],
        "parse_strings_as_datetimes": False,
        "result_format": {
            "result_format": "COMPLETE",
            "partial_unexpected_count": 20,
            "include_unexpected_rows": False,
        },
    }

    engine: PandasExecutionEngine = build_pandas_engine(df=df)
    (
        table_columns_metric,
        unexpected_columns_metric,
        metrics,
    ) = _build_table_columns_and_unexpected(engine, metric_value_kwargs)
    desired_metric = MetricConfiguration(
        metric_name=f"column_values.in_set.{SummarizationMetricNameSuffixes.UNEXPECTED_INDEX_LIST.value}",
        metric_domain_kwargs={"column": "animals"},
        metric_value_kwargs=metric_value_kwargs,
    )
    desired_metric.metric_dependencies = {
        "unexpected_condition": unexpected_columns_metric,
        "table.columns": table_columns_metric,
    }
    results: Dict[Tuple[str, str, str], MetricValue] = engine.resolve_metrics(
        metrics_to_resolve=(desired_metric,), metrics=metrics
    )
    for key, val in results.items():
        assert val == [3, 4, 5]


@pytest.mark.unit
def test_pd_unexpected_index_list_metric_without_id_pk_without_column_values(
    animal_table_df,
):
    df: pd.DataFrame = animal_table_df
    # pandas will return default unexpected_index_list without id_pk
    metric_value_kwargs: dict = {
        "value_set": ["cat", "fish", "dog"],
        "parse_strings_as_datetimes": False,
        "result_format": {
            "result_format": "COMPLETE",
            "partial_unexpected_count": 20,
            "include_unexpected_rows": False,
            "exclude_unexpected_values": True,
        },
    }

    engine: PandasExecutionEngine = build_pandas_engine(df=df)
    (
        table_columns_metric,
        unexpected_columns_metric,
        metrics,
    ) = _build_table_columns_and_unexpected(engine, metric_value_kwargs)
    desired_metric = MetricConfiguration(
        metric_name=f"column_values.in_set.{SummarizationMetricNameSuffixes.UNEXPECTED_INDEX_LIST.value}",
        metric_domain_kwargs={"column": "animals"},
        metric_value_kwargs=metric_value_kwargs,
    )
    desired_metric.metric_dependencies = {
        "unexpected_condition": unexpected_columns_metric,
        "table.columns": table_columns_metric,
    }
    results: Dict[Tuple[str, str, str], MetricValue] = engine.resolve_metrics(
        metrics_to_resolve=(desired_metric,), metrics=metrics
    )
    for key, val in results.items():
        assert val == [3, 4, 5]


@pytest.mark.unit
def test_pd_unexpected_index_list_metric_with_id_pk(metric_value_kwargs_complete, animal_table_df):
    df: pd.DataFrame = animal_table_df
    metric_value_kwargs: dict = metric_value_kwargs_complete

    engine: PandasExecutionEngine = build_pandas_engine(df=df)
    (
        table_columns_metric,
        unexpected_columns_metric,
        metrics,
    ) = _build_table_columns_and_unexpected(engine, metric_value_kwargs_complete)

    unexpected_index_list: MetricConfiguration = MetricConfiguration(
        metric_name=f"column_values.in_set.{SummarizationMetricNameSuffixes.UNEXPECTED_INDEX_LIST.value}",
        metric_domain_kwargs={"column": "animals"},
        metric_value_kwargs=metric_value_kwargs,
    )
    unexpected_index_list.metric_dependencies = {
        "unexpected_condition": unexpected_columns_metric,
        "table.columns": table_columns_metric,
    }
    results: Dict[Tuple[str, str, str], MetricValue] = engine.resolve_metrics(
        metrics_to_resolve=(unexpected_index_list,), metrics=metrics
    )
    for key, val in results.items():
        assert val == [
            {"animals": "giraffe", "pk_1": 3},
            {"animals": "lion", "pk_1": 4},
            {"animals": "zebra", "pk_1": 5},
        ]


@pytest.mark.unit
def test_pd_unexpected_index_list_metric_with_id_pk_without_column_values(
    metric_value_kwargs_complete, animal_table_df
):
    df: pd.DataFrame = animal_table_df
    metric_value_kwargs: dict = {
        "value_set": ["cat", "fish", "dog"],
        "parse_strings_as_datetimes": False,
        "result_format": {
            "result_format": "COMPLETE",
            "unexpected_index_column_names": ["pk_1"],
            "partial_unexpected_count": 20,
            "include_unexpected_rows": False,
            "exclude_unexpected_values": True,
        },
    }

    engine: PandasExecutionEngine = build_pandas_engine(df=df)
    (
        table_columns_metric,
        unexpected_columns_metric,
        metrics,
    ) = _build_table_columns_and_unexpected(engine, metric_value_kwargs_complete)

    unexpected_index_list: MetricConfiguration = MetricConfiguration(
        metric_name=f"column_values.in_set.{SummarizationMetricNameSuffixes.UNEXPECTED_INDEX_LIST.value}",
        metric_domain_kwargs={"column": "animals"},
        metric_value_kwargs=metric_value_kwargs,
    )
    unexpected_index_list.metric_dependencies = {
        "unexpected_condition": unexpected_columns_metric,
        "table.columns": table_columns_metric,
    }
    results: Dict[Tuple[str, str, str], MetricValue] = engine.resolve_metrics(
        metrics_to_resolve=(unexpected_index_list,), metrics=metrics
    )
    for key, val in results.items():
        assert val == [
            {"pk_1": [3, 4, 5]},
        ]


@pytest.mark.unit
def test_sa_unexpected_index_list_metric_with_id_pk(
    sa, animal_table_df, metric_value_kwargs_complete
):
    df: pd.DataFrame = animal_table_df
    metric_value_kwargs: dict = metric_value_kwargs_complete

    engine: SqlAlchemyExecutionEngine = build_sa_execution_engine(df=df, sa=sa)
    (
        table_columns_metric,
        unexpected_columns_metric,
        metrics,
    ) = _build_table_columns_and_unexpected(engine, metric_value_kwargs_complete)

    unexpected_index_list = MetricConfiguration(
        metric_name=f"column_values.in_set.{SummarizationMetricNameSuffixes.UNEXPECTED_INDEX_LIST.value}",
        metric_domain_kwargs={"column": "animals"},
        metric_value_kwargs=metric_value_kwargs,
    )
    unexpected_index_list.metric_dependencies = {
        "unexpected_condition": unexpected_columns_metric,
        "table.columns": table_columns_metric,
    }
    results: Dict[Tuple[str, str, str], MetricValue] = engine.resolve_metrics(
        metrics_to_resolve=(unexpected_index_list,), metrics=metrics
    )
    for key, val in results.items():
        assert val == [
            {"animals": "giraffe", "pk_1": 3},
            {"animals": "lion", "pk_1": 4},
            {"animals": "zebra", "pk_1": 5},
        ]


@pytest.mark.unit
def test_sa_unexpected_index_list_metric_with_id_pk_without_column_values(
    sa, animal_table_df, metric_value_kwargs_complete
):
    df: pd.DataFrame = animal_table_df
    metric_value_kwargs: dict = {
        "value_set": ["cat", "fish", "dog"],
        "parse_strings_as_datetimes": False,
        "result_format": {
            "result_format": "COMPLETE",
            "unexpected_index_column_names": ["pk_1"],
            "partial_unexpected_count": 20,
            "include_unexpected_rows": False,
            "exclude_unexpected_values": True,
        },
    }

    engine: SqlAlchemyExecutionEngine = build_sa_execution_engine(df=df, sa=sa)
    (
        table_columns_metric,
        unexpected_columns_metric,
        metrics,
    ) = _build_table_columns_and_unexpected(engine, metric_value_kwargs_complete)

    unexpected_index_list = MetricConfiguration(
        metric_name=f"column_values.in_set.{SummarizationMetricNameSuffixes.UNEXPECTED_INDEX_LIST.value}",
        metric_domain_kwargs={"column": "animals"},
        metric_value_kwargs=metric_value_kwargs,
    )
    unexpected_index_list.metric_dependencies = {
        "unexpected_condition": unexpected_columns_metric,
        "table.columns": table_columns_metric,
    }
    results: Dict[Tuple[str, str, str], MetricValue] = engine.resolve_metrics(
        metrics_to_resolve=(unexpected_index_list,), metrics=metrics
    )
    for key, val in results.items():
        assert val == [
            {"pk_1": [3, 4, 5]},
        ]


@pytest.mark.unit
def test_sa_unexpected_index_list_metric_without_id_pk(sa, animal_table_df):
    df: pd.DataFrame = animal_table_df
    metric_value_kwargs: dict = {
        "value_set": ["cat", "fish", "dog"],
        "parse_strings_as_datetimes": False,
        "result_format": {
            "result_format": "COMPLETE",
            "partial_unexpected_count": 20,
            "include_unexpected_rows": False,
        },
    }

    engine: SqlAlchemyExecutionEngine = build_sa_execution_engine(df=df, sa=sa)
    (
        table_columns_metric,
        unexpected_columns_metric,
        metrics,
    ) = _build_table_columns_and_unexpected(engine, metric_value_kwargs)

    unexpected_index_list = MetricConfiguration(
        metric_name=f"column_values.in_set.{SummarizationMetricNameSuffixes.UNEXPECTED_INDEX_LIST.value}",
        metric_domain_kwargs={"column": "animals"},
        metric_value_kwargs=metric_value_kwargs,
    )
    unexpected_index_list.metric_dependencies = {
        "unexpected_condition": unexpected_columns_metric,
        "table.columns": table_columns_metric,
    }
    results: Dict[Tuple[str, str, str], MetricValue] = engine.resolve_metrics(
        metrics_to_resolve=(unexpected_index_list,), metrics=metrics
    )
    assert list(results.values())[0] is None


@pytest.mark.unit
def test_sa_unexpected_index_list_metric_without_id_pk_without_column_values(sa, animal_table_df):
    df: pd.DataFrame = animal_table_df
    metric_value_kwargs: dict = {
        "value_set": ["cat", "fish", "dog"],
        "parse_strings_as_datetimes": False,
        "result_format": {
            "result_format": "COMPLETE",
            "partial_unexpected_count": 20,
            "include_unexpected_rows": False,
            "exclude_unexpected_values": True,
        },
    }

    engine: SqlAlchemyExecutionEngine = build_sa_execution_engine(df=df, sa=sa)
    (
        table_columns_metric,
        unexpected_columns_metric,
        metrics,
    ) = _build_table_columns_and_unexpected(engine, metric_value_kwargs)

    unexpected_index_list = MetricConfiguration(
        metric_name=f"column_values.in_set.{SummarizationMetricNameSuffixes.UNEXPECTED_INDEX_LIST.value}",
        metric_domain_kwargs={"column": "animals"},
        metric_value_kwargs=metric_value_kwargs,
    )
    unexpected_index_list.metric_dependencies = {
        "unexpected_condition": unexpected_columns_metric,
        "table.columns": table_columns_metric,
    }
    results: Dict[Tuple[str, str, str], MetricValue] = engine.resolve_metrics(
        metrics_to_resolve=(unexpected_index_list,), metrics=metrics
    )
    assert list(results.values())[0] is None


@pytest.mark.unit
def test_sa_unexpected_index_query_metric_with_id_pk(
    sa, animal_table_df, metric_value_kwargs_complete
):
    df: pd.DataFrame = animal_table_df
    metric_value_kwargs: dict = metric_value_kwargs_complete

    engine = build_sa_execution_engine(df=df, sa=sa)
    (
        table_columns_metric,
        unexpected_columns_metric,
        metrics,
    ) = _build_table_columns_and_unexpected(engine, metric_value_kwargs_complete)
    unexpected_index_query: MetricConfiguration = MetricConfiguration(
        metric_name=f"column_values.in_set.{SummarizationMetricNameSuffixes.UNEXPECTED_INDEX_QUERY.value}",
        metric_domain_kwargs={"column": "animals"},
        metric_value_kwargs=metric_value_kwargs,
    )
    unexpected_index_query.metric_dependencies = {
        "unexpected_condition": unexpected_columns_metric,
        "table.columns": table_columns_metric,
    }
    results: Dict[Tuple[str, str, str], MetricValue] = engine.resolve_metrics(
        metrics_to_resolve=(unexpected_index_query,), metrics=metrics
    )
    for key, val in results.items():
        assert (
            val == "SELECT pk_1, animals \n"
            "FROM test \n"
            "WHERE animals IS NOT NULL AND (animals NOT IN ('cat', 'fish', 'dog'));"
        )


@pytest.mark.unit
def test_sa_unexpected_index_query_metric_without_id_pk(sa, animal_table_df):
    df: pd.DataFrame = animal_table_df
    metric_value_kwargs: dict = {
        "value_set": ["cat", "fish", "dog"],
        "parse_strings_as_datetimes": False,
        "result_format": {
            "result_format": "COMPLETE",
            "partial_unexpected_count": 20,
            "include_unexpected_rows": False,
        },
    }

    engine: SqlAlchemyExecutionEngine = build_sa_execution_engine(df=df, sa=sa)

    (
        table_columns_metric,
        unexpected_columns_metric,
        metrics,
    ) = _build_table_columns_and_unexpected(engine, metric_value_kwargs)
    unexpected_index_query: MetricConfiguration = MetricConfiguration(
        metric_name=f"column_values.in_set.{SummarizationMetricNameSuffixes.UNEXPECTED_INDEX_QUERY.value}",
        metric_domain_kwargs={"column": "animals"},
        metric_value_kwargs=metric_value_kwargs,
    )
    unexpected_index_query.metric_dependencies = {
        "unexpected_condition": unexpected_columns_metric,
        "table.columns": table_columns_metric,
    }
    results: Dict[Tuple[str, str, str], MetricValue] = engine.resolve_metrics(
        metrics_to_resolve=(unexpected_index_query,), metrics=metrics
    )
    for key, val in results.items():
        assert (
            val == "SELECT animals \n"
            "FROM test \n"
            "WHERE animals IS NOT NULL AND (animals NOT IN ('cat', 'fish', 'dog'));"
        )


@pytest.mark.spark
def test_spark_unexpected_index_list_metric_with_id_pk(
    spark_session, animal_table_df, metric_value_kwargs_complete
):
    engine: SparkDFExecutionEngine = build_spark_engine(
        spark=spark_session, df=animal_table_df, batch_id="my_id"
    )

    (
        table_columns_metric,
        unexpected_columns_metric,
        metrics,
    ) = _build_table_columns_and_unexpected(engine, metric_value_kwargs_complete)

    unexpected_index_list = MetricConfiguration(
        metric_name=f"column_values.in_set.{SummarizationMetricNameSuffixes.UNEXPECTED_INDEX_LIST.value}",
        metric_domain_kwargs={"column": "animals"},
        metric_value_kwargs=metric_value_kwargs_complete,
    )
    unexpected_index_list.metric_dependencies = {
        "unexpected_condition": unexpected_columns_metric,
        "table.columns": table_columns_metric,
    }
    results: Dict[Tuple[str, str, str], MetricValue] = engine.resolve_metrics(
        metrics_to_resolve=(unexpected_index_list,), metrics=metrics
    )
    for val in results.values():
        assert val == [
            {"animals": "giraffe", "pk_1": 3},
            {"animals": "lion", "pk_1": 4},
            {"animals": "zebra", "pk_1": 5},
        ]


@pytest.mark.spark
def test_spark_unexpected_index_list_metric_with_id_pk_without_column_values(
    spark_session, animal_table_df
):
    metric_value_kwargs: dict = {
        "value_set": ["cat", "fish", "dog"],
        "parse_strings_as_datetimes": False,
        "result_format": {
            "result_format": "COMPLETE",
            "unexpected_index_column_names": ["pk_1"],
            "partial_unexpected_count": 20,
            "include_unexpected_rows": False,
            "exclude_unexpected_values": True,
        },
    }

    engine: SparkDFExecutionEngine = build_spark_engine(
        spark=spark_session, df=animal_table_df, batch_id="my_id"
    )

    (
        table_columns_metric,
        unexpected_columns_metric,
        metrics,
    ) = _build_table_columns_and_unexpected(engine, metric_value_kwargs)

    unexpected_index_list = MetricConfiguration(
        metric_name=f"column_values.in_set.{SummarizationMetricNameSuffixes.UNEXPECTED_INDEX_LIST.value}",
        metric_domain_kwargs={"column": "animals"},
        metric_value_kwargs=metric_value_kwargs,
    )
    unexpected_index_list.metric_dependencies = {
        "unexpected_condition": unexpected_columns_metric,
        "table.columns": table_columns_metric,
    }
    results: Dict[Tuple[str, str, str], MetricValue] = engine.resolve_metrics(
        metrics_to_resolve=(unexpected_index_list,), metrics=metrics
    )
    for val in results.values():
        assert val == [
            {"pk_1": [3, 4, 5]},
        ]


@pytest.mark.spark
def test_spark_unexpected_index_list_metric_without_id_pk(spark_session, animal_table_df):
    metric_value_kwargs: dict = {
        "value_set": ["cat", "fish", "dog"],
        "parse_strings_as_datetimes": False,
        "result_format": {
            "result_format": "COMPLETE",
            "partial_unexpected_count": 20,
            "include_unexpected_rows": False,
        },
    }

    engine: SparkDFExecutionEngine = build_spark_engine(
        spark=spark_session, df=animal_table_df, batch_id="my_id"
    )
    (
        table_columns_metric,
        unexpected_columns_metric,
        metrics,
    ) = _build_table_columns_and_unexpected(engine, metric_value_kwargs)

    unexpected_index_list = MetricConfiguration(
        metric_name=f"column_values.in_set.{SummarizationMetricNameSuffixes.UNEXPECTED_INDEX_LIST.value}",
        metric_domain_kwargs={"column": "animals"},
        metric_value_kwargs=metric_value_kwargs,
    )
    unexpected_index_list.metric_dependencies = {
        "unexpected_condition": unexpected_columns_metric,
        "table.columns": table_columns_metric,
    }
    results: Dict[Tuple[str, str, str], MetricValue] = engine.resolve_metrics(
        metrics_to_resolve=(unexpected_index_list,), metrics=metrics
    )
    assert list(results.values())[0] is None


@pytest.mark.spark
def test_spark_unexpected_index_list_metric_without_id_pk_without_column_values(
    spark_session, animal_table_df
):
    metric_value_kwargs: dict = {
        "value_set": ["cat", "fish", "dog"],
        "parse_strings_as_datetimes": False,
        "result_format": {
            "result_format": "COMPLETE",
            "partial_unexpected_count": 20,
            "include_unexpected_rows": False,
            "exclude_unexpected_values": True,
        },
    }

    engine: SparkDFExecutionEngine = build_spark_engine(
        spark=spark_session, df=animal_table_df, batch_id="my_id"
    )
    (
        table_columns_metric,
        unexpected_columns_metric,
        metrics,
    ) = _build_table_columns_and_unexpected(engine, metric_value_kwargs)

    unexpected_index_list = MetricConfiguration(
        metric_name=f"column_values.in_set.{SummarizationMetricNameSuffixes.UNEXPECTED_INDEX_LIST.value}",
        metric_domain_kwargs={"column": "animals"},
        metric_value_kwargs=metric_value_kwargs,
    )
    unexpected_index_list.metric_dependencies = {
        "unexpected_condition": unexpected_columns_metric,
        "table.columns": table_columns_metric,
    }
    results: Dict[Tuple[str, str, str], MetricValue] = engine.resolve_metrics(
        metrics_to_resolve=(unexpected_index_list,), metrics=metrics
    )
    assert list(results.values())[0] is None


@pytest.mark.big
def test_pd_unexpected_index_query_metric_with_id_pk(animal_table_df, metric_value_kwargs_complete):
    df: pd.DataFrame = animal_table_df
    metric_value_kwargs: dict = metric_value_kwargs_complete
    engine: PandasExecutionEngine = build_pandas_engine(df=df)

    (
        table_columns_metric,
        unexpected_columns_metric,
        metrics,
    ) = _build_table_columns_and_unexpected(engine, metric_value_kwargs_complete)
    unexpected_index_query: MetricConfiguration = MetricConfiguration(
        metric_name=f"column_values.in_set.{SummarizationMetricNameSuffixes.UNEXPECTED_INDEX_QUERY.value}",
        metric_domain_kwargs={"column": "animals"},
        metric_value_kwargs=metric_value_kwargs,
    )
    unexpected_index_query.metric_dependencies = {
        "unexpected_condition": unexpected_columns_metric,
        "table.columns": table_columns_metric,
    }
    results: Dict[Tuple[str, str, str], MetricValue] = engine.resolve_metrics(
        metrics_to_resolve=(unexpected_index_query,), metrics=metrics
    )
    for val in results.values():
        assert val == "df.filter(items=[3, 4, 5], axis=0)"


@pytest.mark.big
def test_pd_unexpected_index_query_metric_without_id_pk(
    animal_table_df,
):
    df: pd.DataFrame = animal_table_df
    metric_value_kwargs: dict = {
        "value_set": ["cat", "fish", "dog"],
        "parse_strings_as_datetimes": False,
        "result_format": {
            "result_format": "COMPLETE",
            "partial_unexpected_count": 20,
            "include_unexpected_rows": False,
        },
    }
    engine: PandasExecutionEngine = build_pandas_engine(df=df)

    (
        table_columns_metric,
        unexpected_columns_metric,
        metrics,
    ) = _build_table_columns_and_unexpected(engine, metric_value_kwargs)
    unexpected_index_query: MetricConfiguration = MetricConfiguration(
        metric_name=f"column_values.in_set.{SummarizationMetricNameSuffixes.UNEXPECTED_INDEX_QUERY.value}",
        metric_domain_kwargs={"column": "animals"},
        metric_value_kwargs=metric_value_kwargs,
    )
    unexpected_index_query.metric_dependencies = {
        "unexpected_condition": unexpected_columns_metric,
        "table.columns": table_columns_metric,
    }
    results: Dict[Tuple[str, str, str], MetricValue] = engine.resolve_metrics(
        metrics_to_resolve=(unexpected_index_query,), metrics=metrics
    )
    for val in results.values():
        assert val == "df.filter(items=[3, 4, 5], axis=0)"


@pytest.mark.spark
def test_spark_unexpected_index_query_metric_with_id_pk(
    spark_session, animal_table_df, metric_value_kwargs_complete
):
    metric_value_kwargs: dict = metric_value_kwargs_complete

    engine: SparkDFExecutionEngine = build_spark_engine(
        spark=spark_session, df=animal_table_df, batch_id="my_id"
    )
    (
        table_columns_metric,
        unexpected_columns_metric,
        metrics,
    ) = _build_table_columns_and_unexpected(engine, metric_value_kwargs_complete)
    unexpected_index_query: MetricConfiguration = MetricConfiguration(
        metric_name=f"column_values.in_set.{SummarizationMetricNameSuffixes.UNEXPECTED_INDEX_QUERY.value}",
        metric_domain_kwargs={"column": "animals"},
        metric_value_kwargs=metric_value_kwargs,
    )
    unexpected_index_query.metric_dependencies = {
        "unexpected_condition": unexpected_columns_metric,
        "table.columns": table_columns_metric,
    }
    results: Dict[Tuple[str, str, str], MetricValue] = engine.resolve_metrics(
        metrics_to_resolve=(unexpected_index_query,), metrics=metrics
    )
    for val in results.values():
        assert (
            val
            == "df.filter(F.expr((animals IS NOT NULL) AND (NOT (animals IN (cat, fish, dog)))))"
        )


@pytest.mark.spark
def test_spark_unexpected_index_query_metric_without_id_pk(
    spark_session, animal_table_df, metric_value_kwargs_complete
):
    metric_value_kwargs: dict = {
        "value_set": ["cat", "fish", "dog"],
        "parse_strings_as_datetimes": False,
        "result_format": {
            "result_format": "COMPLETE",
            "partial_unexpected_count": 20,
            "include_unexpected_rows": False,
        },
    }

    engine: SparkDFExecutionEngine = build_spark_engine(
        spark=spark_session, df=animal_table_df, batch_id="my_id"
    )

    (
        table_columns_metric,
        unexpected_columns_metric,
        metrics,
    ) = _build_table_columns_and_unexpected(engine, metric_value_kwargs)
    unexpected_index_query: MetricConfiguration = MetricConfiguration(
        metric_name=f"column_values.in_set.{SummarizationMetricNameSuffixes.UNEXPECTED_INDEX_QUERY.value}",
        metric_domain_kwargs={"column": "animals"},
        metric_value_kwargs=metric_value_kwargs,
    )
    unexpected_index_query.metric_dependencies = {
        "unexpected_condition": unexpected_columns_metric,
        "table.columns": table_columns_metric,
    }
    results: Dict[Tuple[str, str, str], MetricValue] = engine.resolve_metrics(
        metrics_to_resolve=(unexpected_index_query,), metrics=metrics
    )
    for val in results.values():
        assert (
            val
            == "df.filter(F.expr((animals IS NOT NULL) AND (NOT (animals IN (cat, fish, dog)))))"
        )



import uuid
from typing import Any, Dict

import pytest
import responses

from great_expectations.data_context.data_context.cloud_data_context import CloudDataContext

CLOUD_BASE_URL = "https://api.greatexpectations.io/fake"
ACCESS_TOKEN = "my-secret-access-token"
ORG_ID = str(uuid.uuid4())
CONTEXT_CONFIGURATION_URL = (
    f"{CLOUD_BASE_URL}/api/v1/organizations/{ORG_ID}/data-context-configuration"
)


def _create_cloud_config_response(
    expectation_suite_store_name_key: str,
    validation_results_store_name_key: str,
    validation_results_store_class_name: str,
) -> Dict[str, Any]:
    return {
        "anonymous_usage_statistics": {
            "data_context_id": "6a52bdfa-e182-455b-a825-e69f076e67d6",
            "enabled": True,
        },
        "checkpoint_store_name": "default_checkpoint_store",
        "config_variables_file_path": "uncommitted/config_variables.yml",
        "config_version": 3.0,
        "data_docs_sites": {},
        expectation_suite_store_name_key: "suite_parameter_store",
        "expectations_store_name": "default_expectations_store",
        "plugins_directory": "plugins/",
        "progress_bars": {
            "globally": False,
            "metric_calculations": False,
            "profilers": False,
        },
        "stores": {
            "default_checkpoint_store": {
                "class_name": "CheckpointStore",
                "store_backend": {
                    "class_name": "GXCloudStoreBackend",
                    "ge_cloud_base_url": CLOUD_BASE_URL,
                    "ge_cloud_credentials": {
                        "access_token": ACCESS_TOKEN,
                        "organization_id": ORG_ID,
                    },
                    "ge_cloud_resource_type": "checkpoint",
                    "suppress_store_backend_id": True,
                },
            },
            "default_expectations_store": {
                "class_name": "ExpectationsStore",
                "store_backend": {
                    "class_name": "GXCloudStoreBackend",
                    "ge_cloud_base_url": CLOUD_BASE_URL,
                    "ge_cloud_credentials": {
                        "access_token": ORG_ID,
                        "organization_id": ORG_ID,
                    },
                    "ge_cloud_resource_type": "expectation_suite",
                    "suppress_store_backend_id": True,
                },
            },
            "default_validation_results_store": {
                "class_name": validation_results_store_class_name,
                "store_backend": {
                    "class_name": "GXCloudStoreBackend",
                    "ge_cloud_base_url": CLOUD_BASE_URL,
                    "ge_cloud_credentials": {
                        "access_token": ACCESS_TOKEN,
                        "organization_id": ORG_ID,
                    },
                    "ge_cloud_resource_type": "validation_result",
                    "suppress_store_backend_id": True,
                },
            },
            "expectations_store": {
                "class_name": "ExpectationsStore",
                "store_backend": {
                    "base_directory": "expectations/",
                    "class_name": "TupleFilesystemStoreBackend",
                },
            },
        },
        validation_results_store_name_key: "default_validation_results_store",
    }


V0_CONFIG = _create_cloud_config_response(
    expectation_suite_store_name_key="evaluation_parameter_store_name",
    validation_results_store_name_key="validations_store_name",
    validation_results_store_class_name="ValidationsStore",
)

V1_CONFIG = _create_cloud_config_response(
    expectation_suite_store_name_key="suite_parameter_store_name",
    validation_results_store_name_key="validation_results_store_name",
    validation_results_store_class_name="ValidationResultsStore",
)


@pytest.mark.parametrize(
    ("config",),
    [
        (V0_CONFIG,),
        (V1_CONFIG,),
    ],
)
@responses.activate
@pytest.mark.unit
def test_parses_v0_config_from_cloud(config: dict):
    """
    Tests to ensure we can build a cloud data context from both v0 and v1 configurations.

    NOTE: This includes some assertions, but we are also just checking that no exceptions
    are raised when instantiating the CloudDataContext, as would happen if we didn't
    properly map keys from the v0 configuration to the v1 configuration.
    """

    responses.add(
        responses.GET,
        CONTEXT_CONFIGURATION_URL,
        json=config,
        status=200,
    )

    CloudDataContext(
        cloud_base_url=CLOUD_BASE_URL,
        cloud_access_token=ACCESS_TOKEN,
        cloud_organization_id=ORG_ID,
    )

    # if we didn't raise when instantiating the context, we are good!



from __future__ import annotations

import warnings

from great_expectations.compatibility.not_imported import NotImported

SPARK_NOT_IMPORTED = NotImported("pyspark is not installed, please 'pip install pyspark'")

with warnings.catch_warnings():
    # DeprecationWarning: typing.io is deprecated, import directly from typing instead. typing.io will be removed in Python 3.12.  # noqa: E501 # FIXME CoP
    warnings.simplefilter(action="ignore", category=DeprecationWarning)
    try:
        import pyspark
    except ImportError:
        pyspark = SPARK_NOT_IMPORTED  # type: ignore[assignment] # FIXME CoP

try:
    from pyspark.sql import functions
except (ImportError, AttributeError):
    functions = SPARK_NOT_IMPORTED  # type: ignore[assignment] # FIXME CoP

try:
    from pyspark.sql import types
except (ImportError, AttributeError):
    types = SPARK_NOT_IMPORTED  # type: ignore[assignment] # FIXME CoP

try:
    from pyspark import SparkContext
except ImportError:
    SparkContext = SPARK_NOT_IMPORTED  # type: ignore[assignment,misc] # FIXME CoP

try:
    from pyspark.ml.feature import Bucketizer
except (ImportError, AttributeError):
    Bucketizer = SPARK_NOT_IMPORTED  # type: ignore[assignment,misc] # FIXME CoP

try:
    from pyspark.sql import Column
except (ImportError, AttributeError):
    Column = SPARK_NOT_IMPORTED  # type: ignore[assignment,misc] # FIXME CoP

try:
    from pyspark.sql.connect.dataframe import DataFrame as ConnectDataFrame
except (ImportError, AttributeError):
    ConnectDataFrame = SPARK_NOT_IMPORTED  # type: ignore[assignment,misc] # FIXME CoP

try:
    from pyspark.sql import DataFrame
except (ImportError, AttributeError):
    DataFrame = SPARK_NOT_IMPORTED  # type: ignore[assignment,misc] # FIXME CoP

try:
    from pyspark.sql import Row
except (ImportError, AttributeError):
    Row = SPARK_NOT_IMPORTED  # type: ignore[assignment,misc] # FIXME CoP

try:
    from pyspark.sql import SparkSession
except (ImportError, AttributeError):
    SparkSession = SPARK_NOT_IMPORTED  # type: ignore[assignment,misc] # FIXME CoP

try:
    from pyspark.sql.connect.session import SparkSession as SparkConnectSession
except (ImportError, AttributeError):
    SparkConnectSession = SPARK_NOT_IMPORTED  # type: ignore[assignment,misc] # FIXME CoP

try:
    from pyspark.sql import SQLContext
except (ImportError, AttributeError):
    SQLContext = SPARK_NOT_IMPORTED  # type: ignore[assignment,misc] # FIXME CoP

try:
    from pyspark.sql import Window
except (ImportError, AttributeError):
    Window = SPARK_NOT_IMPORTED  # type: ignore[assignment,misc] # FIXME CoP

try:
    from pyspark.sql.readwriter import DataFrameReader
except (ImportError, AttributeError):
    DataFrameReader = SPARK_NOT_IMPORTED  # type: ignore[assignment,misc] # FIXME CoP

try:
    from pyspark.sql.utils import AnalysisException
except (ImportError, AttributeError):
    AnalysisException = SPARK_NOT_IMPORTED  # type: ignore[assignment,misc] # FIXME CoP

try:
    from pyspark.errors import PySparkAttributeError
except (ImportError, AttributeError):
    PySparkAttributeError = SPARK_NOT_IMPORTED  # type: ignore[assignment,misc] # FIXME CoP



from typing import List, Literal, Union, get_args

from great_expectations.compatibility.pydantic import Field
from great_expectations.compatibility.typing_extensions import Annotated
from great_expectations.core.suite_parameters import (
    SuiteParameterDict,  # used in pydantic validation
)
from great_expectations.expectations.model_field_descriptions import (
    MOSTLY_DESCRIPTION,
    VALUE_SET_DESCRIPTION,
)

MostlyField = Annotated[
    Union[float, SuiteParameterDict],
    Field(
        description=MOSTLY_DESCRIPTION,
        ge=0.0,
        le=1.0,
        # This is just for the schema, it should not be validated on input
        schema_overrides={"multipleOf": 0.01},
    ),
]

ValueSetField = Annotated[
    Union[List, SuiteParameterDict],
    Field(
        title="Value Set",
        description=VALUE_SET_DESCRIPTION,
        schema_overrides={
            "anyOf": [
                {
                    "title": "Value Set",
                    "description": VALUE_SET_DESCRIPTION,
                    "oneOf": [
                        {
                            "title": "Text",
                            "type": "array",
                            "items": {"type": "string", "minLength": 1},
                            "minItems": 1,
                            "examples": [
                                ["a", "b", "c", "d", "e"],
                                [
                                    "2024-01-01",
                                    "2024-01-02",
                                    "2024-01-03",
                                    "2024-01-04",
                                    "2024-01-05",
                                ],
                            ],
                        },
                        {
                            "title": "Numbers",
                            "type": "array",
                            "items": {"type": "number"},
                            "minItems": 1,
                            "examples": [
                                [1, 2, 3, 4, 5],
                                [1.1, 2.2, 3.3, 4.4, 5.5],
                                [1, 2.2, 3, 4.4, 5],
                            ],
                        },
                    ],
                },
                {"type": "object"},
            ]
        },
    ),
]


# If you re-order these strings you must reorder the variable names from the `get_args`
# call below as well or the variables will reference the wrong string
ConditionParser = Literal[
    "great_expectations", "great_expectations__experimental__", "pandas", "spark"
]

(
    CONDITION_PARSER_GREAT_EXPECTATIONS,
    CONDITION_PARSER_GREAT_EXPECTATIONS_DEPRECATED,
    CONDITION_PARSER_PANDAS,
    CONDITION_PARSER_SPARK,
) = get_args(ConditionParser)



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Dict, List, Optional

from great_expectations.core.metric_domain_types import MetricDomainTypes
from great_expectations.execution_engine import (
    SparkDFExecutionEngine,
    SqlAlchemyExecutionEngine,
)
from great_expectations.expectations.metrics.metric_provider import metric_value
from great_expectations.expectations.metrics.query_metric_provider import (
    QueryMetricProvider,
    QueryParameters,
)

if TYPE_CHECKING:
    from great_expectations.compatibility import pyspark


class QueryColumnPair(QueryMetricProvider):
    metric_name = "query.column_pair"
    value_keys = (
        "column_A",
        "column_B",
        "query",
    )

    @metric_value(engine=SqlAlchemyExecutionEngine)
    def _sqlalchemy(
        cls,
        execution_engine: SqlAlchemyExecutionEngine,
        metric_domain_kwargs: dict,
        metric_value_kwargs: dict,
        metrics: Dict[str, Any],
        runtime_configuration: dict,
    ) -> list[dict]:
        batch_selectable, _, _ = execution_engine.get_compute_domain(
            metric_domain_kwargs, domain_type=MetricDomainTypes.TABLE
        )
        query = cls._get_query_from_metric_value_kwargs(metric_value_kwargs)
        column_A: Optional[str] = metric_value_kwargs.get("column_A")
        column_B: Optional[str] = metric_value_kwargs.get("column_B")
        if column_A and column_B:
            query_parameters = QueryParameters(
                column_A=column_A,
                column_B=column_B,
            )
        else:
            raise ValueError("Both `column_A` and `column_B` must be provided.")  # noqa: TRY003 # FIXME CoP
        substituted_batch_subquery = (
            cls._get_substituted_batch_subquery_from_query_and_batch_selectable(
                query=query,
                batch_selectable=batch_selectable,
                execution_engine=execution_engine,
                query_parameters=query_parameters,
            )
        )
        return cls._get_sqlalchemy_records_from_substituted_batch_subquery(
            substituted_batch_subquery=substituted_batch_subquery,
            execution_engine=execution_engine,
        )

    @metric_value(engine=SparkDFExecutionEngine)
    def _spark(
        cls,
        execution_engine: SparkDFExecutionEngine,
        metric_domain_kwargs: dict,
        metric_value_kwargs: dict,
        metrics: Dict[str, Any],
        runtime_configuration: dict,
    ) -> List[dict]:
        query = cls._get_query_from_metric_value_kwargs(metric_value_kwargs)

        df: pyspark.DataFrame
        df, _, _ = execution_engine.get_compute_domain(
            metric_domain_kwargs, domain_type=MetricDomainTypes.TABLE
        )

        df.createOrReplaceTempView("tmp_view")
        column_A: Optional[str] = metric_value_kwargs.get("column_A")
        column_B: Optional[str] = metric_value_kwargs.get("column_B")
        query = query.format(column_A=column_A, column_B=column_B, batch="tmp_view")

        engine: pyspark.SparkSession = execution_engine.spark
        result: List[pyspark.Row] = engine.sql(query).collect()

        return [element.asDict() for element in result]



import os

import pytest

import great_expectations as gx
from great_expectations.data_context import CloudDataContext


class V1GetContextError(Exception):
    pass


@pytest.mark.cloud
@pytest.mark.parametrize("analytics_enabled", [True, False])
def test_get_context(monkeypatch, analytics_enabled):
    from great_expectations.analytics.config import ENV_CONFIG

    monkeypatch.setattr(ENV_CONFIG, "gx_analytics_enabled", analytics_enabled)
    context = gx.get_context(
        mode="cloud",
        cloud_base_url=os.environ.get("GX_CLOUD_BASE_URL"),
        cloud_organization_id=os.environ.get("GX_CLOUD_ORGANIZATION_ID"),
        cloud_access_token=os.environ.get("GX_CLOUD_ACCESS_TOKEN"),
    )
    assert isinstance(context, CloudDataContext)
    # This assert is to ensure we are hitting the v1 and not the v0 endpoint.
    assert context.config.config_version == 4.0



from datetime import datetime, timezone
from typing import Sequence

import pandas as pd

import great_expectations.expectations as gxe
from tests.integration.conftest import parameterize_batch_for_data_sources
from tests.integration.test_utils.data_source_config import (
    BigQueryDatasourceTestConfig,
    DatabricksDatasourceTestConfig,
    DataSourceTestConfig,
    MSSQLDatasourceTestConfig,
    MySQLDatasourceTestConfig,
    PandasDataFrameDatasourceTestConfig,
    PandasFilesystemCsvDatasourceTestConfig,
    PostgreSQLDatasourceTestConfig,
    RedshiftDatasourceTestConfig,
    SnowflakeDatasourceTestConfig,
    SparkFilesystemCsvDatasourceTestConfig,
    SqliteDatasourceTestConfig,
)

ALL_DATA_SOURCES: Sequence[DataSourceTestConfig] = [
    BigQueryDatasourceTestConfig(),
    DatabricksDatasourceTestConfig(),
    MSSQLDatasourceTestConfig(),
    MySQLDatasourceTestConfig(),
    PandasDataFrameDatasourceTestConfig(),
    PandasFilesystemCsvDatasourceTestConfig(),
    PostgreSQLDatasourceTestConfig(),
    RedshiftDatasourceTestConfig(),
    SnowflakeDatasourceTestConfig(),
    SparkFilesystemCsvDatasourceTestConfig(),
    SqliteDatasourceTestConfig(),
]

NON_SQL_DATA_SOURCES: Sequence[DataSourceTestConfig] = [
    PandasDataFrameDatasourceTestConfig(),
    PandasFilesystemCsvDatasourceTestConfig(),
    SparkFilesystemCsvDatasourceTestConfig(),
]

SQL_DATA_SOURCES: Sequence[DataSourceTestConfig] = [
    BigQueryDatasourceTestConfig(),
    DatabricksDatasourceTestConfig(),
    MSSQLDatasourceTestConfig(),
    MySQLDatasourceTestConfig(),
    PostgreSQLDatasourceTestConfig(),
    RedshiftDatasourceTestConfig(),
    SnowflakeDatasourceTestConfig(),
    SqliteDatasourceTestConfig(),
]

DATA_SOURCES_THAT_SUPPORT_DATE_COMPARISONS: Sequence[DataSourceTestConfig] = [
    BigQueryDatasourceTestConfig(),
    DatabricksDatasourceTestConfig(),
    MSSQLDatasourceTestConfig(),
    MySQLDatasourceTestConfig(),
    PandasDataFrameDatasourceTestConfig(),
    PostgreSQLDatasourceTestConfig(),
    RedshiftDatasourceTestConfig(),
    SnowflakeDatasourceTestConfig(),
    SparkFilesystemCsvDatasourceTestConfig(),
]

JUST_PANDAS_DATA_SOURCES: Sequence[DataSourceTestConfig] = [PandasDataFrameDatasourceTestConfig()]


@parameterize_batch_for_data_sources(
    data_source_configs=ALL_DATA_SOURCES,
    data=pd.DataFrame({"a": [1, 2]}),
)
def test_expect_column_min_to_be_between(batch_for_datasource) -> None:
    expectation = gxe.ExpectColumnMinToBeBetween(column="a", min_value=1, max_value=1)
    result = batch_for_datasource.validate(expectation)
    assert result.success


@parameterize_batch_for_data_sources(
    data_source_configs=ALL_DATA_SOURCES,
    data=pd.DataFrame(
        {
            "date": [
                datetime(year=2021, month=1, day=31, tzinfo=timezone.utc).date(),
                datetime(year=2022, month=1, day=31, tzinfo=timezone.utc).date(),
                datetime(year=2023, month=1, day=31, tzinfo=timezone.utc).date(),
            ]
        }
    ),
)
def test_expect_column_min_to_be_between__date(batch_for_datasource) -> None:
    expectation = gxe.ExpectColumnMinToBeBetween(
        column="date",
        min_value=datetime(year=2021, month=1, day=1, tzinfo=timezone.utc).date(),
        max_value=datetime(year=2022, month=1, day=1, tzinfo=timezone.utc).date(),
    )
    result = batch_for_datasource.validate(expectation)
    assert result.success


@parameterize_batch_for_data_sources(
    data_source_configs=ALL_DATA_SOURCES,
    data=pd.DataFrame(
        {
            "date": [
                datetime(year=2021, month=1, day=31, tzinfo=timezone.utc).date(),
                datetime(year=2022, month=1, day=31, tzinfo=timezone.utc).date(),
                datetime(year=2023, month=1, day=31, tzinfo=timezone.utc).date(),
            ]
        }
    ),
)
def test_expect_column_max_to_be_between__date(batch_for_datasource) -> None:
    expectation = gxe.ExpectColumnMaxToBeBetween(
        column="date",
        min_value=datetime(year=2023, month=1, day=1, tzinfo=timezone.utc).date(),
        max_value=datetime(year=2024, month=1, day=1, tzinfo=timezone.utc).date(),
    )
    result = batch_for_datasource.validate(expectation)
    assert result.success


@parameterize_batch_for_data_sources(
    data_source_configs=ALL_DATA_SOURCES,
    data=pd.DataFrame({"a": [1, 2]}),
)
def test_expect_column_max_to_be_between(batch_for_datasource) -> None:
    expectation = gxe.ExpectColumnMaxToBeBetween(column="a", min_value=2, max_value=2)
    result = batch_for_datasource.validate(expectation)
    assert result.success



from __future__ import annotations

import logging
import warnings
from pprint import pformat as pf
from typing import (
    TYPE_CHECKING,
    Any,
    ClassVar,
    Dict,
    Generic,
    List,
    Literal,
    Optional,
    Type,
    TypeVar,
    Union,
)

import great_expectations.exceptions as gx_exceptions
from great_expectations._docs_decorators import (
    public_api,
)
from great_expectations.compatibility import pydantic
from great_expectations.compatibility.pydantic import (
    StrictBool,
    StrictFloat,
    StrictInt,
    StrictStr,
)
from great_expectations.compatibility.pyspark import ConnectDataFrame, DataFrame, pyspark
from great_expectations.compatibility.typing_extensions import override
from great_expectations.core import IDDict
from great_expectations.core.batch import LegacyBatchDefinition
from great_expectations.core.batch_spec import RuntimeDataBatchSpec
from great_expectations.datasource.fluent import BatchParameters, BatchRequest
from great_expectations.datasource.fluent.batch_identifier_util import make_batch_identifier
from great_expectations.datasource.fluent.constants import (
    _DATA_CONNECTOR_NAME,
)
from great_expectations.datasource.fluent.interfaces import (
    Batch,
    DataAsset,
    Datasource,
    TestConnectionError,
    _DataAssetT,
)
from great_expectations.exceptions.exceptions import BuildBatchRequestError

if TYPE_CHECKING:
    from typing_extensions import TypeAlias, TypeGuard

    from great_expectations.compatibility.pyspark import SparkSession
    from great_expectations.core.batch_definition import BatchDefinition
    from great_expectations.core.partitioners import ColumnPartitioner
    from great_expectations.datasource.fluent.data_connector.batch_filter import BatchSlice
    from great_expectations.datasource.fluent.interfaces import BatchMetadata
    from great_expectations.execution_engine import SparkDFExecutionEngine

logger = logging.getLogger(__name__)


# this enables us to include dataframe in the json schema
_SparkDataFrameT = TypeVar("_SparkDataFrameT")

SparkConfig: TypeAlias = Dict[StrictStr, Union[StrictStr, StrictInt, StrictFloat, StrictBool]]


class SparkDatasourceError(Exception):
    pass


class _SparkDatasource(Datasource):
    # instance attributes
    spark_config: Union[SparkConfig, None] = None
    force_reuse_spark_context: bool = True
    persist: bool = True

    # private attrs
    _spark: Union[SparkSession, None] = pydantic.PrivateAttr(None)

    @pydantic.validator("force_reuse_spark_context")
    @classmethod
    def _force_reuse_spark_context_deprecation_warning(cls, v: bool) -> bool:
        if v is not None:
            # deprecated-v1.0.0
            warnings.warn(
                "force_reuse_spark_context is deprecated and will be removed in version 1.0. "
                "In environments that allow it, the existing Spark context will be reused, adding the "  # noqa: E501 # FIXME CoP
                "spark_config options that have been passed. If the Spark context cannot be updated with "  # noqa: E501 # FIXME CoP
                "the spark_config, the context will be stopped and restarted with the new spark_config.",  # noqa: E501 # FIXME CoP
                category=DeprecationWarning,
            )
        return v

    @classmethod
    @override
    def update_forward_refs(cls) -> None:  # type: ignore[override] # FIXME CoP
        from great_expectations.compatibility.pyspark import SparkSession

        super().update_forward_refs(SparkSession=SparkSession)

    @staticmethod
    @override
    def _update_asset_forward_refs(asset_type: Type[_DataAssetT]) -> None:
        # Only update forward refs if pyspark types are available
        if pyspark:
            asset_type.update_forward_refs()

    # Abstract Methods
    @property
    @override
    def execution_engine_type(self) -> Type[SparkDFExecutionEngine]:
        """Return the SparkDFExecutionEngine unless the override is set"""
        from great_expectations.execution_engine.sparkdf_execution_engine import (
            SparkDFExecutionEngine,
        )

        return SparkDFExecutionEngine

    def get_spark(self) -> SparkSession:
        # circular imports require us to import SparkSession and update_forward_refs
        # only when assigning to self._spark for SparkSession isinstance check
        self.update_forward_refs()
        self._spark: SparkSession = self.execution_engine_type.get_or_create_spark_session(
            spark_config=self.spark_config,
        )
        return self._spark

    @override
    def get_execution_engine(self) -> SparkDFExecutionEngine:
        # Method override is required because PrivateAttr _spark won't be passed into Execution Engine  # noqa: E501 # FIXME CoP
        # unless it is passed explicitly.
        current_execution_engine_kwargs = self.dict(
            exclude=self._get_exec_engine_excludes(),
            config_provider=self._config_provider,
        )
        if (
            current_execution_engine_kwargs != self._cached_execution_engine_kwargs
            or not self._execution_engine
        ):
            if self._spark:
                self._execution_engine = self._execution_engine_type()(
                    spark=self._spark, **current_execution_engine_kwargs
                )
            else:
                self._execution_engine = self._execution_engine_type()(
                    **current_execution_engine_kwargs
                )

            self._cached_execution_engine_kwargs = current_execution_engine_kwargs
        return self._execution_engine

    @override
    def test_connection(self, test_assets: bool = True) -> None:
        """Test the connection for the _SparkDatasource.

        Args:
            test_assets: If assets have been passed to the _SparkDatasource,
                         an attempt can be made to test them as well.

        Raises:
            TestConnectionError: If the connection test fails.
        """
        try:
            self.get_spark()
        except Exception as e:
            raise TestConnectionError(cause=e) from e

    # End Abstract Methods


@public_api
class DataFrameAsset(DataAsset, Generic[_SparkDataFrameT]):
    """
    A DataAsset that represents a Spark DataFrame.
    """

    # instance attributes
    type: Literal["dataframe"] = "dataframe"

    class Config:
        extra = pydantic.Extra.forbid

    @override
    def test_connection(self) -> None: ...

    @override
    def get_batch_parameters_keys(
        self, partitioner: Optional[ColumnPartitioner] = None
    ) -> tuple[str, ...]:
        return tuple(
            "dataframe",
        )

    def _get_reader_method(self) -> str:
        raise NotImplementedError(
            """Spark DataFrameAsset does not implement "_get_reader_method()" method, because DataFrame is already available."""  # noqa: E501 # FIXME CoP
        )

    def _get_reader_options_include(self) -> set[str]:
        raise NotImplementedError(
            """Spark DataFrameAsset does not implement "_get_reader_options_include()" method, because DataFrame is already available."""  # noqa: E501 # FIXME CoP
        )

    @override
    def build_batch_request(
        self,
        options: Optional[BatchParameters] = None,
        batch_slice: Optional[BatchSlice] = None,
        partitioner: Optional[ColumnPartitioner] = None,
    ) -> BatchRequest:
        """A batch request that can be used to obtain batches for this DataAsset.

        Args:
            options: This should have 1 key, 'dataframe', whose value is the datafame to validate.
            batch_slice: This is not currently supported and must be None for this data asset.
            partitioner: This is not currently supported and must be None for this data asset.


        Returns:
            A BatchRequest object that can be used to obtain a batch from an Asset by calling the
            get_batch method.
        """
        if batch_slice is not None:
            raise BuildBatchRequestError(
                message="batch_slice is not currently supported for this DataAsset "
                "and must be None."
            )

        if partitioner is not None:
            raise BuildBatchRequestError(
                message="partitioner is not currently supported for this DataAsset "
                "and must be None."
            )

        if not (options is not None and "dataframe" in options and len(options) == 1):
            raise BuildBatchRequestError(message="options must contain exactly 1 key, 'dataframe'.")

        if not self.is_spark_data_frame(options["dataframe"]):
            raise BuildBatchRequestError(
                message="Cannot build batch request without a Spark DataFrame."
            )

        return BatchRequest(
            datasource_name=self.datasource.name,
            data_asset_name=self.name,
            options=options,
        )

    @override
    def _validate_batch_request(self, batch_request: BatchRequest) -> None:
        """Validates the batch_request has the correct form.

        Args:
            batch_request: A batch request object to be validated.
        """
        if not (
            batch_request.datasource_name == self.datasource.name
            and batch_request.data_asset_name == self.name
            and batch_request.options
            and len(batch_request.options) == 1
            and "dataframe" in batch_request.options
            and self.is_spark_data_frame(batch_request.options["dataframe"])
        ):
            expect_batch_request_form = BatchRequest[None](
                datasource_name=self.datasource.name,
                data_asset_name=self.name,
                options={},
                batch_slice=batch_request._batch_slice_input,  # type: ignore[attr-defined] # FIXME CoP
            )
            raise gx_exceptions.InvalidBatchRequestError(  # noqa: TRY003 # FIXME CoP
                "BatchRequest should have form:\n"
                f"{pf(expect_batch_request_form.dict())}\n"
                f"but actually has form:\n{pf(batch_request.dict())}\n"
            )

    @override
    def get_batch_identifiers_list(self, batch_request: BatchRequest) -> List[dict]:
        return [IDDict(batch_request.options)]

    @override
    def get_batch(self, batch_request: BatchRequest) -> Batch:
        self._validate_batch_request(batch_request)

        batch_spec = RuntimeDataBatchSpec(batch_data=batch_request.options["dataframe"])
        execution_engine: SparkDFExecutionEngine = self.datasource.get_execution_engine()
        data, markers = execution_engine.get_batch_data_and_markers(batch_spec=batch_spec)

        # batch_definition (along with batch_spec and markers) is only here to satisfy a
        # legacy constraint when computing usage statistics in a validator. We hope to remove
        # it in the future.
        batch_definition = LegacyBatchDefinition(
            datasource_name=self.datasource.name,
            data_connector_name=_DATA_CONNECTOR_NAME,
            data_asset_name=self.name,
            batch_identifiers=make_batch_identifier(batch_request.options),
            batch_spec_passthrough=None,
        )

        batch_metadata: BatchMetadata = self._get_batch_metadata_from_batch_request(
            batch_request=batch_request, ignore_options=("dataframe",)
        )

        return Batch(
            datasource=self.datasource,
            data_asset=self,
            batch_request=batch_request,
            data=data,
            metadata=batch_metadata,
            batch_markers=markers,
            batch_spec=batch_spec,
            batch_definition=batch_definition,
        )

    @public_api
    def add_batch_definition_whole_dataframe(self, name: str) -> BatchDefinition:
        """
        Add a BatchDefinition that represents the entire DataFrame.

        Args:
            name: The name of the Batch Definition.

        Returns:
            A BatchDefinition object that represents the entire DataFrame.
        """
        return self.add_batch_definition(
            name=name,
            partitioner=None,
        )

    @staticmethod
    def is_spark_data_frame(df: Any) -> TypeGuard[Union[DataFrame, ConnectDataFrame]]:
        """Check that a given object is a Spark DataFrame.
        This could either be a regular Spark DataFrame or a Spark Connect DataFrame.
        """
        data_frame_types = [DataFrame, ConnectDataFrame]
        return any((cls and isinstance(df, cls)) for cls in data_frame_types)


@public_api
class SparkDatasource(_SparkDatasource):
    """
    A SparkDatasource is a Datasource that connects to a Spark cluster and provides
    access to Spark DataFrames.
    """

    # class attributes
    asset_types: ClassVar[List[Type[DataAsset]]] = [DataFrameAsset]

    # instance attributes
    type: Literal["spark"] = "spark"

    assets: List[DataFrameAsset] = []

    @public_api
    def add_dataframe_asset(
        self,
        name: str,
        batch_metadata: Optional[BatchMetadata] = None,
    ) -> DataFrameAsset:
        """Adds a Dataframe DataAsset to this SparkDatasource object.

        Args:
            name: The name of the DataFrame asset. This can be any arbitrary string.
            dataframe: The Spark Dataframe containing the data for this DataFrame data asset.
            batch_metadata: An arbitrary user defined dictionary with string keys which will get inherited by any
                            batches created from the asset.

        Returns:
            The DataFrameAsset that has been added to this datasource.
        """  # noqa: E501 # FIXME CoP
        asset: DataFrameAsset = DataFrameAsset(
            name=name,
            batch_metadata=batch_metadata or {},
        )
        return self._add_asset(asset=asset)



from __future__ import annotations

import datetime as dt
import json
from typing import (
    TYPE_CHECKING,
    AbstractSet,
    Any,
    Callable,
    Dict,
    List,
    Mapping,
    Optional,
    TypedDict,
    Union,
    cast,
)

import great_expectations.exceptions as gx_exceptions
from great_expectations._docs_decorators import public_api
from great_expectations.analytics import submit as submit_analytics_event
from great_expectations.analytics.events import CheckpointRanEvent
from great_expectations.checkpoint.actions import (
    _VALIDATION_ACTION_REGISTRY,
    ActionContext,
    UpdateDataDocsAction,
    ValidationAction,
)
from great_expectations.compatibility.pydantic import (
    BaseModel,
    Extra,
    Field,
    root_validator,
    validator,
)
from great_expectations.compatibility.typing_extensions import override
from great_expectations.core.expectation_validation_result import (
    ExpectationSuiteValidationResult,
)
from great_expectations.core.freshness_diagnostics import CheckpointFreshnessDiagnostics
from great_expectations.core.result_format import DEFAULT_RESULT_FORMAT, ResultFormatUnion
from great_expectations.core.run_identifier import RunIdentifier
from great_expectations.core.serdes import _IdentifierBundle
from great_expectations.core.suite_parameters import SuiteParameterDict
from great_expectations.core.validation_definition import ValidationDefinition
from great_expectations.data_context.data_context.context_factory import project_manager
from great_expectations.data_context.types.resource_identifiers import (
    ExpectationSuiteIdentifier,
    ValidationResultIdentifier,
)
from great_expectations.exceptions import (
    CheckpointNotAddedError,
    CheckpointNotFreshError,
    CheckpointRunWithoutValidationDefinitionError,
)
from great_expectations.exceptions.exceptions import (
    CheckpointNotFoundError,
    InvalidKeyError,
    StoreBackendError,
)
from great_expectations.exceptions.resource_freshness import ResourceFreshnessAggregateError
from great_expectations.render.renderer.renderer import Renderer

if TYPE_CHECKING:
    from great_expectations.data_context.store.validation_definition_store import (
        ValidationDefinitionStore,
    )


@public_api
class Checkpoint(BaseModel):
    """
    A Checkpoint is the primary means for validating data in a production deployment of Great Expectations.

    Checkpoints provide a convenient abstraction for running a number of validation definitions and triggering a set of actions
    to be taken after the validation step.

    Args:
        name: The name of the checkpoint.
        validation_definitions: List of validation definitions to be run.
        actions: List of actions to be taken after the validation definitions are run.
        result_format: The format in which to return the results of the validation definitions. Default is ResultFormat.SUMMARY.
        id: An optional unique identifier for the checkpoint.

    """  # noqa: E501 # FIXME CoP

    name: str
    validation_definitions: List[ValidationDefinition]
    actions: List[ValidationAction] = Field(default_factory=list)
    result_format: ResultFormatUnion = DEFAULT_RESULT_FORMAT
    id: Union[str, None] = None

    class Config:
        """
        When serialized, the validation_definitions field will be encoded as a set of identifiers.
        These will be used as foreign keys to retrieve the actual objects from the appropriate stores.

        Example:
        {
            "name": "my_checkpoint",
            "validation_definitions": [
                {
                    "name": "my_first_validation",
                    "id": "a58816-64c8-46cb-8f7e-03c12cea1d67"
                },
                {
                    "name": "my_second_validation",
                    "id": "139ab16-64c8-46cb-8f7e-03c12cea1d67"
                },
            ],
            "actions": [
                {
                    "name": "my_slack_action",
                    "slack_webhook": "https://hooks.slack.com/services/ABC123/DEF456/XYZ789",
                    "notify_on": "all",
                    "notify_with": ["my_data_docs_site"],
                    "renderer": {
                        "class_name": "SlackRenderer",
                    }
                }
            ],
            "result_format": "SUMMARY",
            "id": "b758816-64c8-46cb-8f7e-03c12cea1d67"
        }
        """  # noqa: E501 # FIXME CoP

        extra = Extra.forbid
        arbitrary_types_allowed = (
            True  # Necessary for compatibility with ValidationAction's Marshmallow dep
        )
        json_encoders = {
            ValidationDefinition: lambda v: v.identifier_bundle(),
            Renderer: lambda r: r.serialize(),
        }

    @validator("actions", pre=True)
    @classmethod
    def validate_actions(
        cls, action_list: list[ValidationAction] | list[dict]
    ) -> list[ValidationAction]:
        validated_actions: list[ValidationAction] = []
        for action in action_list:
            if isinstance(action, ValidationAction):
                validated_actions.append(action)
            else:
                action_type: str | None = action.get("type")
                action_cls = _VALIDATION_ACTION_REGISTRY.get(action_type)
                validated_action = action_cls(**action)
                validated_actions.append(validated_action)

        return validated_actions

    @override
    def json(  # noqa: PLR0913 # FIXME CoP
        self,
        *,
        include: AbstractSet[int | str] | Mapping[int | str, Any] | None = None,
        exclude: AbstractSet[int | str] | Mapping[int | str, Any] | None = None,
        by_alias: bool = False,
        skip_defaults: bool | None = None,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
        encoder: Callable[[Any], Any] | None = None,
        models_as_dict: bool = True,
        **dumps_kwargs: Any,
    ) -> str:
        """
        Override the default json method to enable proper diagnostics around validation_definitions.

        NOTE: This should be removed in favor of a field/model serializer when we upgrade to
              Pydantic 2.
        """
        json_data = super().json(
            include=include,
            exclude=self._determine_exclude(exclude=exclude),
            by_alias=by_alias,
            skip_defaults=skip_defaults,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
            encoder=encoder,
            models_as_dict=models_as_dict,
        )

        data = json.loads(json_data)  # Parse back to dict to add validation_definitions
        data_with_validation_definitions = self._serialize_validation_definitions(data)
        return json.dumps(data_with_validation_definitions, **dumps_kwargs)

    @override
    def dict(  # noqa: PLR0913 # FIXME CoP
        self,
        *,
        include: AbstractSet[int | str] | Mapping[int | str, Any] | None = None,
        exclude: AbstractSet[int | str] | Mapping[int | str, Any] | None = None,
        by_alias: bool = False,
        skip_defaults: bool | None = None,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
    ) -> Dict[str, Any]:
        """
        Override the default dict method to enable proper diagnostics around validation_definitions.

        NOTE: This should be removed in favor of a field/model serializer when we upgrade to
              Pydantic 2.
        """
        data = super().dict(
            include=include,
            exclude=self._determine_exclude(exclude=exclude),
            by_alias=by_alias,
            skip_defaults=skip_defaults,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
        )

        return self._serialize_validation_definitions(data=data)

    def _determine_exclude(
        self, exclude: AbstractSet[int | str] | Mapping[int | str, Any] | None
    ) -> AbstractSet[int | str] | Mapping[int | str, Any] | None:
        if not exclude:
            exclude = set()

        if isinstance(exclude, set):
            exclude.add("validation_definitions")
        else:
            exclude["__all__"] = "validation_definitions"  # type: ignore[index] # FIXME

        return exclude

    def _serialize_validation_definitions(self, data: Dict[str, Any]) -> Dict[str, Any]:
        """
        Manually serialize the validation_definitions field to avoid Pydantic's default
        serialization.

        We want to aggregate all errors from the validation_definitions and raise them as
        a single error.
        """
        data["validation_definitions"] = []

        diagnostics = CheckpointFreshnessDiagnostics(errors=[])
        for validation_definition in self.validation_definitions:
            try:
                identifier_bundle = validation_definition.identifier_bundle()
                data["validation_definitions"].append(identifier_bundle.dict())
            except ResourceFreshnessAggregateError as e:
                diagnostics.errors.extend(e.errors)

        diagnostics.raise_for_error()
        return data

    @validator("validation_definitions", pre=True)
    def _validate_validation_definitions(
        cls, validation_definitions: list[ValidationDefinition] | list[Dict[str, Any]]
    ) -> list[ValidationDefinition]:
        if validation_definitions and isinstance(validation_definitions[0], Dict):
            validation_definition_store = project_manager.get_validation_definition_store()
            identifier_bundles = [
                _IdentifierBundle(**v)  # type: ignore[arg-type] # All validation configs are dicts if the first one is
                for v in validation_definitions
            ]
            return cls._deserialize_identifier_bundles_to_validation_definitions(
                identifier_bundles=identifier_bundles, store=validation_definition_store
            )

        return cast("List[ValidationDefinition]", validation_definitions)

    @classmethod
    def _deserialize_identifier_bundles_to_validation_definitions(
        cls, identifier_bundles: list[_IdentifierBundle], store: ValidationDefinitionStore
    ) -> list[ValidationDefinition]:
        validation_definitions: list[ValidationDefinition] = []
        for id_bundle in identifier_bundles:
            key = store.get_key(name=id_bundle.name, id=id_bundle.id)

            try:
                validation_definition = store.get(key=key)
            except (KeyError, gx_exceptions.InvalidKeyError):
                raise ValueError(f"Unable to retrieve validation definition {id_bundle} from store")  # noqa: TRY003 # FIXME CoP

            if not validation_definition:
                raise ValueError(  # noqa: TRY003 # FIXME CoP
                    "ValidationDefinitionStore did not retrieve a validation definition"
                )
            validation_definitions.append(validation_definition)

        return validation_definitions

    @public_api
    def run(
        self,
        batch_parameters: Dict[str, Any] | None = None,
        expectation_parameters: SuiteParameterDict | None = None,
        run_id: RunIdentifier | None = None,
    ) -> CheckpointResult:
        """
        Runs the Checkpoint's underlying Validation Definitions and Actions.

        Args:
            batch_parameters: Parameters to be used when loading the Batch.
            expectation_parameters: Parameters to be used when validating the Batch.
            run_id: An optional unique identifier for the run.

        Returns:
            A CheckpointResult object containing the results of the run.

        Raises:
            CheckpointRunWithoutValidationDefinitionError: If the Checkpoint is run without any
                                                           Validation Definitions.
            CheckpointNotAddedError: If the Checkpoint has not been added to the Store.
            CheckpointNotFreshError: If the Checkpoint has been modified since it was last added
                                     to the Store.
        """
        if not self.validation_definitions:
            raise CheckpointRunWithoutValidationDefinitionError()

        diagnostics = self.is_fresh()
        if not diagnostics.success:
            # The checkpoint itself is not added but all children are - we can add it for the user
            if not diagnostics.parent_added and diagnostics.children_added:
                self._add_to_store()
            else:
                diagnostics.raise_for_error()

        if batch_parameters is None:
            batch_parameters = {}
        if expectation_parameters is None:
            expectation_parameters = SuiteParameterDict()
        run_id = run_id or RunIdentifier(run_time=dt.datetime.now(dt.timezone.utc))
        self._prepare_checkpoint_run_for_context(batch_parameters, expectation_parameters)
        run_results = self._run_validation_definitions(
            batch_parameters=batch_parameters,
            expectation_parameters=expectation_parameters,
            result_format=self.result_format,
            run_id=run_id,
        )

        checkpoint_result = self._construct_result(run_id=run_id, run_results=run_results)
        self._run_actions(checkpoint_result=checkpoint_result)

        self._submit_analytics_event()

        return checkpoint_result

    def _prepare_checkpoint_run_for_context(
        self,
        batch_parameters: Dict[str, Any],
        expectation_parameters: SuiteParameterDict,
    ) -> None:
        context = self.validation_definitions[0].data.data_asset.datasource.data_context
        context.prepare_checkpoint_run(self, batch_parameters, expectation_parameters)

    def _submit_analytics_event(self):
        event = CheckpointRanEvent(
            checkpoint_id=self.id,
            validation_definition_ids=[val_def.id for val_def in self.validation_definitions],
        )
        submit_analytics_event(event=event)

    def _run_validation_definitions(
        self,
        batch_parameters: Dict[str, Any],
        expectation_parameters: SuiteParameterDict,
        result_format: ResultFormatUnion,
        run_id: RunIdentifier,
    ) -> Dict[ValidationResultIdentifier, ExpectationSuiteValidationResult]:
        run_results: Dict[ValidationResultIdentifier, ExpectationSuiteValidationResult] = {}
        for validation_definition in self.validation_definitions:
            validation_result = validation_definition.run(
                checkpoint_id=self.id,
                batch_parameters=batch_parameters,
                expectation_parameters=expectation_parameters,
                result_format=result_format,
                run_id=run_id,
            )
            key = self._build_result_key(
                validation_definition=validation_definition,
                run_id=run_id,
                batch_identifier=validation_result.batch_id,
            )
            run_results[key] = validation_result

        return run_results

    def _build_result_key(
        self,
        validation_definition: ValidationDefinition,
        run_id: RunIdentifier,
        batch_identifier: Optional[str] = None,
    ) -> ValidationResultIdentifier:
        return ValidationResultIdentifier(
            expectation_suite_identifier=ExpectationSuiteIdentifier(
                name=validation_definition.suite.name
            ),
            run_id=run_id,
            batch_identifier=batch_identifier,
        )

    def _construct_result(
        self,
        run_id: RunIdentifier,
        run_results: Dict[ValidationResultIdentifier, ExpectationSuiteValidationResult],
    ) -> CheckpointResult:
        for result in run_results.values():
            result.meta["checkpoint_id"] = self.id

        return CheckpointResult(
            run_id=run_id,
            run_results=run_results,
            checkpoint_config=self,
        )

    def _run_actions(
        self,
        checkpoint_result: CheckpointResult,
    ) -> None:
        action_context = ActionContext()
        sorted_actions = self._sort_actions()
        for action in sorted_actions:
            action_result = action.run(
                checkpoint_result=checkpoint_result,
                action_context=action_context,
            )
            action_context.update(action=action, action_result=action_result)

    def _sort_actions(self) -> List[ValidationAction]:
        """
        UpdateDataDocsActions are prioritized to run first, followed by all other actions.

        This is due to the fact that certain actions reference data docs sites,
        which must be updated first.
        """
        priority_actions: List[ValidationAction] = []
        secondary_actions: List[ValidationAction] = []
        for action in self.actions:
            if isinstance(action, UpdateDataDocsAction):
                priority_actions.append(action)
            else:
                secondary_actions.append(action)

        return priority_actions + secondary_actions

    def is_fresh(self) -> CheckpointFreshnessDiagnostics:
        checkpoint_diagnostics = CheckpointFreshnessDiagnostics(
            errors=[] if self.id else [CheckpointNotAddedError(name=self.name)]
        )
        validation_definition_diagnostics = [vd.is_fresh() for vd in self.validation_definitions]
        checkpoint_diagnostics.update_with_children(*validation_definition_diagnostics)

        if not checkpoint_diagnostics.success:
            return checkpoint_diagnostics

        store = project_manager.get_checkpoints_store()
        key = store.get_key(name=self.name, id=self.id)

        try:
            checkpoint = store.get(key=key)
        except (
            StoreBackendError,  # Generic error from stores
            InvalidKeyError,  # Ephemeral context error
        ):
            return CheckpointFreshnessDiagnostics(errors=[CheckpointNotFoundError(name=self.name)])

        return CheckpointFreshnessDiagnostics(
            errors=[] if checkpoint == self else [CheckpointNotFreshError(name=self.name)]
        )

    @public_api
    def save(self) -> None:
        """Save the current state of this Checkpoint."""
        store = project_manager.get_checkpoints_store()
        key = store.get_key(name=self.name, id=self.id)

        store.update(key=key, value=self)

    def _add_to_store(self) -> None:
        """This is used to persist a checkpoint before we run it.

        We need to persist a checkpoint before it can be run. If user calls runs but hasn't
        persisted it we add it for them.
        """
        store = project_manager.get_checkpoints_store()
        key = store.get_key(name=self.name, id=self.id)

        store.add(key=key, value=self)


@public_api
class CheckpointResult(BaseModel):
    """
    The result of running a Checkpoint.

    Contains information about Expectation successes and failures from running
    each Validation Definition in the Checkpoint.
    """

    run_id: RunIdentifier
    run_results: Dict[ValidationResultIdentifier, ExpectationSuiteValidationResult]
    checkpoint_config: Checkpoint
    success: Optional[bool] = None

    class Config:
        extra = Extra.forbid
        arbitrary_types_allowed = True

    @root_validator
    def _root_validate_result(cls, values: dict) -> dict:
        run_results = values["run_results"]
        if len(run_results) == 0:
            raise ValueError("CheckpointResult must contain at least one run result")  # noqa: TRY003 # FIXME CoP

        if values["success"] is None:
            values["success"] = all(result.success for result in run_results.values())
        return values

    @property
    def name(self) -> str:
        return self.checkpoint_config.name

    def describe_dict(self) -> CheckpointDescriptionDict:
        success_count = sum(1 for r in self.run_results.values() if r.success)
        run_result_descriptions = [r.describe_dict() for r in self.run_results.values()]
        num_results = len(run_result_descriptions)

        return {
            "success": success_count == num_results,
            "statistics": {
                "evaluated_validations": num_results,
                "success_percent": success_count / num_results * 100,
                "successful_validations": success_count,
                "unsuccessful_validations": num_results - success_count,
            },
            "validation_results": run_result_descriptions,
        }

    @public_api
    def describe(self) -> str:
        """JSON string description of this CheckpointResult"""
        return json.dumps(self.describe_dict(), indent=4)


# Necessary due to cyclic dependencies between Checkpoint and CheckpointResult
CheckpointResult.update_forward_refs()


class CheckpointDescriptionDict(TypedDict):
    success: bool
    statistics: CheckpointDescriptionStatistics
    validation_results: List[Dict[str, Any]]


class CheckpointDescriptionStatistics(TypedDict):
    evaluated_validations: int
    success_percent: float
    successful_validations: int
    unsuccessful_validations: int



from __future__ import annotations

import pathlib
from typing import TYPE_CHECKING, Callable, Final

import pact
import pytest

from tests.integration.cloud.rest_contracts.conftest import (
    EXISTING_ORGANIZATION_ID,
    ContractInteraction,
)

if TYPE_CHECKING:
    from tests.integration.cloud.rest_contracts.conftest import PactBody

NON_EXISTENT_CHECKPOINT_ID: Final[str] = "6ed9a340-8469-4ee2-a300-ffbe5d09b49d"

EXISTING_CHECKPOINT_ID: Final[str] = "4d734009-2c50-4222-bd71-9660f5b05fff"

# Get Checkpoint
GET_CHECKPOINT_MIN_CHECKPOINT_BODY: Final[PactBody] = {
    "id": pact.Format().uuid,
    "type": "checkpoint",
    "attributes": {
        "id": pact.Format().uuid,
        "name": pact.Like("string checkpoint name"),
        "organization_id": pact.Format().uuid,
        "checkpoint_config": {},
    },
}

GET_CHECKPOINT_MIN_RESPONSE_BODY: Final[PactBody] = {
    "data": pact.Like(GET_CHECKPOINT_MIN_CHECKPOINT_BODY)
}
GET_CHECKPOINT_NOT_FOUND_RESPONSE_BODY: Final[PactBody] = pact.Like("404 string")

# Get Checkpoints
GET_CHECKPOINTS_MIN_RESPONSE_BODY: Final[PactBody] = {
    "data": pact.EachLike(
        GET_CHECKPOINT_MIN_CHECKPOINT_BODY,
        minimum=1,  # Default but writing it out for clarity
    ),
}


@pytest.mark.cloud
@pytest.mark.parametrize(
    "contract_interaction",
    [
        ContractInteraction(
            method="GET",
            request_path=pathlib.Path(
                "/",
                "organizations",
                EXISTING_ORGANIZATION_ID,
                "checkpoints",
                EXISTING_CHECKPOINT_ID,
            ),
            upon_receiving="a request to get a Checkpoint",
            given="the Checkpoint exists",
            response_status=200,
            response_body=GET_CHECKPOINT_MIN_RESPONSE_BODY,
        ),
        ContractInteraction(
            method="GET",
            request_path=pathlib.Path(
                "/",
                "organizations",
                EXISTING_ORGANIZATION_ID,
                "checkpoints",
                NON_EXISTENT_CHECKPOINT_ID,
            ),
            upon_receiving="a request to get a Checkpoint",
            given="the Checkpoint does not exist",
            response_status=404,
            response_body=GET_CHECKPOINT_NOT_FOUND_RESPONSE_BODY,
        ),
        ContractInteraction(
            method="GET",
            request_path=pathlib.Path(
                "/",
                "organizations",
                EXISTING_ORGANIZATION_ID,
                "checkpoints",
            ),
            upon_receiving="a request to get all Checkpoints",
            given="the Checkpoints exist",
            response_status=200,
            response_body=GET_CHECKPOINTS_MIN_RESPONSE_BODY,
        ),
    ],
)
def test_checkpoint(
    contract_interaction: ContractInteraction,
    run_rest_api_pact_test: Callable[[ContractInteraction], None],
) -> None:
    run_rest_api_pact_test(contract_interaction)



from __future__ import annotations

from typing import Callable, List, Union

import pytest
from pytest import MonkeyPatch

from great_expectations.compatibility import pydantic
from great_expectations.core.config_provider import (
    _ConfigurationProvider,
    _EnvironmentConfigurationProvider,
)
from great_expectations.datasource.fluent.config_str import (
    ConfigStr,
    ConfigUri,
    SecretStr,
)
from great_expectations.datasource.fluent.fluent_base_model import FluentBaseModel
from great_expectations.exceptions import MissingConfigVariableError

pytestmark = pytest.mark.unit


class MyClass(FluentBaseModel):
    normal_field: str
    secret_field: SecretStr
    config_field: ConfigStr
    config_field_w_default: ConfigStr = r"hey-${MY_SECRET}"  # type: ignore[assignment] # FIXME CoP


@pytest.fixture
def env_config_provider() -> _ConfigurationProvider:
    config_provider = _ConfigurationProvider()
    config_provider.register_provider(_EnvironmentConfigurationProvider())
    return config_provider


def test_config_provider_substitution(
    monkeypatch: MonkeyPatch, env_config_provider: _ConfigurationProvider
):
    monkeypatch.setenv("MY_CONFIG", "bar")

    my_dict = {"my_key": r"foo${MY_CONFIG}", "another_key": r"${MY_CONFIG}"}
    subbed = env_config_provider.substitute_config(my_dict)

    assert subbed == {"my_key": "foobar", "another_key": "bar"}
    assert subbed != my_dict


def test_config_provider_substitution_raises_error(
    monkeypatch: MonkeyPatch, env_config_provider: _ConfigurationProvider
):
    monkeypatch.setenv("NOT_MY_CONFIG", "bar")

    my_dict = {"my_key": r"foo${MY_CONFIG}", "another_key": r"${MY_CONFIG}"}

    with pytest.raises(MissingConfigVariableError):
        env_config_provider.substitute_config(my_dict)


def test_config_str_validation():
    with pytest.raises(pydantic.ValidationError, match="ConfigStr"):
        m = MyClass(normal_field="normal", secret_field="secret", config_field="invalid config")
        print(m)


@pytest.mark.parametrize(
    ["input_value", "expected"],
    [
        (r"${VALID_CONFIG_STR}", r"${VALID_CONFIG_STR}"),
        (
            r"postgres://user:${VALID_CONFIG_STR}@host/dbname",
            r"postgres://user:${VALID_CONFIG_STR}@host/dbname",
        ),
        ("postgres://userName@hostname", "postgres://userName@hostname"),
    ],
)
def test_as_union_file_type(input_value, expected):
    class MyClass(FluentBaseModel):
        my_field: Union[ConfigStr, pydantic.networks.PostgresDsn]

    m = MyClass(my_field=input_value)
    print(m)

    assert str(m.my_field) == expected


def test_config_substitution(monkeypatch: MonkeyPatch, env_config_provider: _ConfigurationProvider):
    monkeypatch.setenv("MY_ENV_VAR", "success")

    m = MyClass(
        normal_field="normal",
        secret_field="secret",  # type: ignore[arg-type] # FIXME CoP
        config_field=r"${MY_ENV_VAR}",  # type: ignore[arg-type] # FIXME CoP
        config_field_w_default=r"hello-${MY_ENV_VAR}",  # type: ignore[arg-type] # FIXME CoP
    )
    assert m.config_field.get_config_value(env_config_provider) == "success"
    assert m.config_field_w_default.get_config_value(env_config_provider) == "hello-success"


def test_config_substitution_dict(
    monkeypatch: MonkeyPatch, env_config_provider: _ConfigurationProvider
):
    monkeypatch.setenv("MY_ENV_VAR", "success")

    m = MyClass(
        normal_field="normal",
        secret_field="secret",  # type: ignore[arg-type] # FIXME CoP
        config_field=r"${MY_ENV_VAR}",  # type: ignore[arg-type] # FIXME CoP
    )

    d = m.dict(config_provider=env_config_provider)
    assert d["config_field"] == "success"


def test_config_nested_substitution_dict(
    monkeypatch: MonkeyPatch, env_config_provider: _ConfigurationProvider
):
    monkeypatch.setenv("MY_ENV_VAR", "success")

    class MyCollection(FluentBaseModel):
        my_classes: List[MyClass] = []

    MyCollection.update_forward_refs(MyClass=MyClass)

    m = MyCollection(
        my_classes=[
            MyClass(
                normal_field="normal",
                secret_field="secret",  # type: ignore[arg-type] # FIXME CoP
                config_field=r"${MY_ENV_VAR}",  # type: ignore[arg-type] # FIXME CoP
            )
        ]
    )

    d = m.dict(config_provider=env_config_provider)
    assert d["my_classes"][0]["config_field"] == "success"


def test_config_nested_substitution_dict_raises_error_for_missing_config_var(
    monkeypatch: MonkeyPatch, env_config_provider: _ConfigurationProvider
):
    monkeypatch.setenv("NOT_MY_ENV_VAR", "failure")

    class MyCollection(FluentBaseModel):
        my_classes: List[MyClass] = []

    MyCollection.update_forward_refs(MyClass=MyClass)

    m = MyCollection(
        my_classes=[
            MyClass(
                normal_field="normal",
                secret_field="secret",  # type: ignore[arg-type] # FIXME CoP
                config_field=r"${MY_ENV_VAR}",  # type: ignore[arg-type] # FIXME CoP
            )
        ]
    )

    with pytest.raises(MissingConfigVariableError):
        m.dict(config_provider=env_config_provider)


@pytest.mark.parametrize("method", ["yaml", "dict", "json"])
def test_serialization_returns_original(monkeypatch: MonkeyPatch, method: str):
    monkeypatch.setenv("MY_ENV_VAR", "dont_serialize_me")

    m = MyClass(
        normal_field="normal",
        secret_field="secret",  # type: ignore[arg-type] # FIXME CoP
        config_field=r"${MY_ENV_VAR}",  # type: ignore[arg-type] # FIXME CoP
    )
    serialize_method: Callable = getattr(m, method)
    dumped = str(serialize_method())
    assert "dont_serialize_me" not in dumped
    assert r"${MY_ENV_VAR}" in dumped


@pytest.mark.parametrize("method", ["yaml", "dict", "json"])
def test_nested_serialization_returns_original(monkeypatch: MonkeyPatch, method: str):
    # TODO: fix this
    class MyCollection(FluentBaseModel):
        my_classes: List[MyClass] = []

    MyCollection.update_forward_refs(MyClass=MyClass)

    monkeypatch.setenv("MY_ENV_VAR", "dont_serialize_me")

    m = MyCollection(
        my_classes=[
            MyClass(
                normal_field="normal",
                secret_field="secret",  # type: ignore[arg-type] # FIXME CoP
                config_field=r"${MY_ENV_VAR}",  # type: ignore[arg-type] # FIXME CoP
            )
        ]
    )
    serialize_method: Callable = getattr(m, method)
    dumped = str(serialize_method())
    assert "dont_serialize_me" not in dumped
    assert r"${MY_ENV_VAR}" in dumped


class TestSecretMasking:
    def test_validation_error(self):
        class MyUnionFields(FluentBaseModel):
            config_field: ConfigStr
            dsn_field: pydantic.PostgresDsn
            union_field: Union[ConfigStr, pydantic.PostgresDsn]

        with pytest.raises(pydantic.ValidationError) as exc_info:
            m = MyUnionFields(
                config_field="invalid_config",
                dsn_field="postgress://:invalid_config@localhost",
                union_field="invalid_config",
            )
            print(m)
        assert "invalid_config" not in str(exc_info.value)

    def test_repr(self):
        m = MyClass(
            normal_field="normal",
            secret_field="secret",
            config_field=r"${MY_ENV_VAR}",
        )
        assert repr(m.config_field) == r"ConfigStr('${MY_ENV_VAR}')"

    def test_str(self):
        m = MyClass(
            normal_field="normal",
            secret_field="secret",
            config_field=r"${MY_ENV_VAR}",
        )
        assert str(m.config_field) == r"${MY_ENV_VAR}"

    def test_serialization(
        self, monkeypatch: MonkeyPatch, env_config_provider: _ConfigurationProvider
    ):
        monkeypatch.setenv("MY_SECRET", "dont_serialize_me")
        m = MyClass(
            normal_field="normal",
            secret_field="my_secret",  # type: ignore[arg-type] # FIXME CoP
            config_field=r"${MY_SECRET}",  # type: ignore[arg-type] # FIXME CoP
        )

        # but it should not actually be used
        for dumped_str in [str(m.dict()), m.yaml(), m.json()]:
            print(dumped_str, "\n\n")
            assert "my_secret" not in dumped_str
            assert "dont_serialize_me" not in dumped_str
            assert r"${MY_SECRET}" in dumped_str


@pytest.mark.parametrize(
    "uri",
    [
        "http://my_user:${MY_PW}@example.com:8000/the/path/?query=here#fragment=is;this=bit",
        "http://${MY_USER}:${MY_PW}@example.com:8000/the/path/?query=here#fragment=is;this=bit",
        "http://${MY_USER}:${MY_PW}@example.com:8000/the/path/?query=here&query=again&query=and_again",
        "snowflake://my_user:${MY_PW}@account/db",
        "snowflake://${MY_USER}:${MY_PW}@account/db",
        "postgresql+psycopg2://my_user:${MY_PW}@host/db",
        "postgresql+psycopg2://${MY_USER}:${MY_PW}@host/db",
    ],
)
class TestConfigUri:
    def test_parts(
        self,
        env_config_provider: _ConfigurationProvider,
        monkeypatch: MonkeyPatch,
        uri: str,
    ):
        monkeypatch.setenv("MY_USER", "my_user")
        monkeypatch.setenv("MY_PW", "super_secret")

        parsed = pydantic.parse_obj_as(ConfigUri, uri)

        # ensure attributes are set
        assert parsed.scheme
        assert parsed.host
        assert parsed.path
        # ensure no attribute errors
        _ = parsed.query
        _ = parsed.fragment

        # ensure that the password (and user) are not substituted
        assert parsed.password == "${MY_PW}"
        assert parsed.user in ["${MY_USER}", "my_user"]

    def test_substitution(
        self,
        env_config_provider: _ConfigurationProvider,
        monkeypatch: MonkeyPatch,
        uri: str,
    ):
        monkeypatch.setenv("MY_USER", "my_user")
        monkeypatch.setenv("MY_PW", "super_secret")

        parsed = pydantic.parse_obj_as(ConfigUri, uri)

        substituted = parsed.get_config_value(env_config_provider)

        # ensure attributes are set
        assert substituted.scheme
        assert substituted.host
        assert substituted.path
        # ensure no attribute errors
        _ = substituted.query
        _ = substituted.fragment

        # ensure that the password (and user) are not substituted
        assert substituted.password == "super_secret"
        assert substituted.user == "my_user"

    def test_leakage(
        self,
        env_config_provider: _ConfigurationProvider,
        monkeypatch: MonkeyPatch,
        uri: str,
    ):
        """
        Ensure the config values are not leaked in the repr or str of the object
        or the component parts.
        """
        monkeypatch.setenv("MY_USER", "my_user")
        monkeypatch.setenv("MY_PW", "super_secret")

        parsed = pydantic.parse_obj_as(ConfigUri, uri)
        assert "super_secret" not in str(parsed)
        assert "super_secret" not in repr(parsed)
        assert parsed.password
        assert "super_secret" not in parsed.password

        if "my_user" not in uri:
            assert "my_user" not in str(parsed)
            assert "my_user" not in repr(parsed)
            assert parsed.user
            assert "my_user" not in parsed.user

    def test_params(self, uri: str):
        parsed = pydantic.parse_obj_as(ConfigUri, uri)
        assert isinstance(parsed.params, dict)

        if parsed.query:
            for q_key, values in parsed.params.items():
                for value in values:
                    assert f"{q_key}={value}" in parsed.query


class TestConfigUriInvalid:
    @pytest.mark.parametrize("uri", ["invalid_uri", "http:/example.com"])
    def test_invalid_uri(self, uri: str):
        with pytest.raises(pydantic.ValidationError):
            _ = pydantic.parse_obj_as(ConfigUri, uri)

    @pytest.mark.parametrize(
        "uri",
        [
            "${MY_SCHEME}://me:secret@account/db/schema",
            "snowflake://me:secret@${MY_ACCOUNT}/db/schema",
            "snowflake://me:secret@account/${MY_DB}/schema",
            "snowflake://me:secret@account/db/${MY_SCHEMA}",
            "snowflake://me:secret@account/db/my_schema?${MY_QUERY_PARAMS}",
            "snowflake://me:secret@account/db/my_schema?role=${MY_ROLE}",
        ],
    )
    def test_disallowed_substitution(self, uri: str):
        with pytest.raises(pydantic.ValidationError):
            _ = pydantic.parse_obj_as(ConfigUri, uri)

    def test_no_template_str(self):
        with pytest.raises(pydantic.ValidationError):
            _ = pydantic.parse_obj_as(ConfigUri, "snowflake://me:password@account/db")


if __name__ == "__main__":
    pytest.main([__file__, "-vv"])



import logging
from typing import List, Optional

from great_expectations.checkpoint import Checkpoint, LegacyCheckpoint

logger = logging.getLogger(__name__)


class ExtendedCheckpoint(Checkpoint):
    def __init__(
        self,
        name: str,
        data_context,
        expectation_suite_name: Optional[str] = None,
        action_list: Optional[List[dict]] = None,
    ):
        super().__init__(
            name=name,
            data_context=data_context,
            expectation_suite_name=expectation_suite_name,
            action_list=action_list,
        )


class ExtendedLegacyCheckpoint(LegacyCheckpoint):
    def __init__(
        self,
        name: str,
        data_context,
        expectation_suite_name: Optional[str] = None,
        action_list: Optional[List[dict]] = None,
    ):
        super().__init__(
            name=name,
            data_context=data_context,
            expectation_suite_name=expectation_suite_name,
            action_list=action_list,
        )


class ExtendedCheckpointIllegalBaseClass:
    def __init__(
        self,
        name: str,
        data_context,
        expectation_suite_name: Optional[str] = None,
        action_list: Optional[List[dict]] = None,
    ):
        super().__init__(
            name=name,
            data_context=data_context,
            expectation_suite_name=expectation_suite_name,
            action_list=action_list,
        )



from typing import Optional

from great_expectations.compatibility.sqlalchemy import sqlalchemy as sa
from great_expectations.compatibility.typing_extensions import override
from great_expectations.core.metric_function_types import SummarizationMetricNameSuffixes
from great_expectations.execution_engine.execution_engine import ExecutionEngine
from great_expectations.execution_engine.sqlalchemy_execution_engine import (
    SqlAlchemyExecutionEngine,
)
from great_expectations.expectations.expectation_configuration import ExpectationConfiguration
from great_expectations.expectations.metrics.column_aggregate_metric_provider import (
    column_aggregate_partial,
)
from great_expectations.expectations.metrics.column_aggregate_metrics.column_standard_deviation import (  # noqa: E501
    ColumnStandardDeviation as BaseColumnStandardDeviation,
)
from great_expectations.validator.metric_configuration import MetricConfiguration


class SqliteExecutionEngine(SqlAlchemyExecutionEngine):
    """SqlAlchemyExecutionEngine for SQLite databases."""

    pass


class ColumnStandardDeviation(BaseColumnStandardDeviation):
    """MetricProvider Class for Aggregate Standard Deviation metric for SQLite databases."""

    # We should change this decorator to compute this metric a completely new way
    @column_aggregate_partial(engine=SqliteExecutionEngine)
    def _sqlalchemy(cls, column, _dialect, _metrics, **kwargs):
        """Sqlite Standard Deviation implementation"""
        mean = _metrics["column.mean"]
        nonnull_row_count = _metrics[
            f"column_values.null.{SummarizationMetricNameSuffixes.UNEXPECTED_COUNT.value}"
        ]
        standard_deviation = sa.func.sqrt(
            sa.func.sum((1.0 * column - mean) * (1.0 * column - mean))
            / ((1.0 * nonnull_row_count) - 1.0)
        )
        return standard_deviation

    @classmethod
    @override
    def _get_evaluation_dependencies(
        cls,
        metric: MetricConfiguration,
        configuration: Optional[ExpectationConfiguration] = None,
        execution_engine: Optional[ExecutionEngine] = None,
        runtime_configuration: Optional[dict] = None,
    ):
        # We don't need to override this here but I wanted to show for completeness
        # If we are changing the decorator on the provider method or we are completely
        # implementing a new datasource, we'll want to override.
        return super()._get_evaluation_dependencies(
            metric=metric,
            configuration=configuration,
            execution_engine=execution_engine,
            runtime_configuration=runtime_configuration,
        )



from unittest import mock

import pytest

from great_expectations.data_context import EphemeralDataContext

SETTER_METHOD_NAMES = [
    "expectations_store_name",
    "validation_results_store_name",
    "checkpoint_store_name",
]


@pytest.mark.unit
@pytest.mark.parametrize(
    "store_setter_method_name",
    SETTER_METHOD_NAMES,
)
def test_store_name_setters(
    store_setter_method_name: str,
    ephemeral_context_with_defaults: EphemeralDataContext,
):
    new_store_name = "new_store_name"
    setattr(ephemeral_context_with_defaults, store_setter_method_name, new_store_name)
    assert getattr(ephemeral_context_with_defaults, store_setter_method_name) == new_store_name


@pytest.mark.unit
@pytest.mark.parametrize(
    "store_setter_method_name",
    SETTER_METHOD_NAMES,
)
def test_store_name_setters_persist(
    store_setter_method_name: str,
    ephemeral_context_with_defaults: EphemeralDataContext,
):
    new_store_name = "new_store_name"
    with mock.patch(
        "great_expectations.data_context.EphemeralDataContext._save_project_config"
    ) as mock_save_project_config:
        setattr(ephemeral_context_with_defaults, store_setter_method_name, new_store_name)

    mock_save_project_config.assert_called_once()



from __future__ import annotations

import pathlib
import re
from typing import TYPE_CHECKING, cast

import boto3
import botocore
import pytest

from great_expectations.core.util import DBFSPath
from great_expectations.datasource.fluent.data_connector import (
    DBFSDataConnector,
)
from great_expectations.datasource.fluent.data_connector.file_path_data_connector import (
    MissingFilePathTemplateMapFnError,
)
from tests.test_utils import create_files_in_directory

if TYPE_CHECKING:
    from pyfakefs.fake_filesystem import FakeFilesystem

    from great_expectations.datasource.fluent.data_connector import (
        DataConnector,
    )


@pytest.mark.filesystem
@pytest.mark.slow  # 1.05s
def test__get_full_file_path_pandas(fs: FakeFilesystem):
    """
    What does this test and why?
    File paths in DBFS need to use the `dbfs:/` protocol base instead of `/dbfs/` when
    being read using the `spark.read` method in the ExecutionEngine.  HOWEVER, when using a
    PandasExecutionEngine the file semantic `/dbfs/` version must be used instead.
    This test verifies that a config using a `/dbfs/` path is NOT translated to `dbfs:/`
    when preparing the PathBatchSpec for the PandasExecutionEngine.
    """
    # Copy boto modules into fake filesystem (see https://github.com/spulec/moto/issues/1682#issuecomment-645016188)
    for module in [boto3, botocore]:
        module_dir = pathlib.Path(module.__file__).parent
        fs.add_real_directory(module_dir, lazy_read=False)

    base_directory: str = "/dbfs/great_expectations"
    fs.create_dir(base_directory)

    create_files_in_directory(
        directory=base_directory,
        file_name_list=[
            "test_dir_0/A/B/C/logfile_0.csv",
            "test_dir_0/A/B/C/bigfile_1.csv",
            "test_dir_0/A/filename2.csv",
            "test_dir_0/A/filename3.csv",
        ],
    )

    my_data_connector = DBFSDataConnector(
        datasource_name="my_file_path_datasource",
        data_asset_name="my_filesystem_data_asset",
        base_directory=pathlib.Path(f"{base_directory}/test_dir_0/A/B/C"),
        glob_directive="*.csv",
        file_path_template_map_fn=DBFSPath.convert_to_file_semantics_version,
    )

    assert (
        my_data_connector._get_full_file_path(path="bigfile_1.csv")
        == f"{base_directory}/test_dir_0/A/B/C/bigfile_1.csv"
    )

    assert my_data_connector.get_data_reference_count() == 2
    assert sorted(my_data_connector.get_data_references()[:3]) == [
        "bigfile_1.csv",
        "logfile_0.csv",
    ]

    batching_regex = re.compile(r"(?P<name>.+)_(?P<number>\d+)\.csv")
    matching_data_references = my_data_connector.get_matched_data_references(regex=batching_regex)
    assert len(matching_data_references) == 2
    assert sorted(matching_data_references[:3]) == [
        "bigfile_1.csv",
        "logfile_0.csv",
    ]


@pytest.mark.unit
def test_dbfs_data_connector_missing_file_path_template_map_fn_error():
    """Test DBFS data connector raises error when file_path_template_map_fn is None."""
    data_connector = DBFSDataConnector(
        datasource_name="my_dbfs_datasource",
        data_asset_name="my_data_asset",
        base_directory=pathlib.Path("/dbfs/test"),
        glob_directive="*.csv",
        file_path_template_map_fn=None,
    )

    with pytest.raises(MissingFilePathTemplateMapFnError):
        data_connector._get_full_file_path("test.csv")


@pytest.mark.spark
def test__get_full_file_path_spark(basic_spark_df_execution_engine, fs):
    """
    What does this test and why?
    File paths in DBFS need to use the `dbfs:/` protocol base instead of `/dbfs/` when
    being read using the `spark.read` method in the ExecutionEngine. In the data connector
    config however, the `/dbfs` version must be used. This test verifies that a config
    using a `/dbfs/` path is translated to `dbfs:/` when preparing the PathBatchSpec for the
    SparkDFExecutionEngine.
    """
    base_directory: str = "/dbfs/great_expectations"
    base_directory_colon: str = "dbfs:/great_expectations"
    fs.create_dir(base_directory)

    create_files_in_directory(
        directory=base_directory,
        file_name_list=[
            "test_dir_0/A/B/C/logfile_0.csv",
            "test_dir_0/A/B/C/bigfile_1.csv",
            "test_dir_0/A/filename2.csv",
            "test_dir_0/A/filename3.csv",
        ],
    )

    my_data_connector: DataConnector = DBFSDataConnector(
        datasource_name="my_file_path_datasource",
        data_asset_name="my_filesystem_data_asset",
        base_directory=pathlib.Path(f"{base_directory}/test_dir_0/A/B/C"),
        glob_directive="*.csv",
        file_path_template_map_fn=DBFSPath.convert_to_protocol_version,
    )

    assert (
        cast("DBFSDataConnector", my_data_connector)._get_full_file_path(path="bigfile_1.csv")
        == f"{base_directory_colon}/test_dir_0/A/B/C/bigfile_1.csv"
    )

    assert my_data_connector.get_data_reference_count() == 2
    assert sorted(my_data_connector.get_data_references()[:3]) == [
        "bigfile_1.csv",
        "logfile_0.csv",
    ]

    batching_regex = re.compile(r"(?P<name>.+)_(?P<number>\d+)\.csv")
    matching_data_references = my_data_connector.get_matched_data_references(regex=batching_regex)
    assert len(matching_data_references) == 2
    assert sorted(matching_data_references[:3]) == [
        "bigfile_1.csv",
        "logfile_0.csv",
    ]



from typing import Any

from great_expectations.metrics.column_pair.column_pair import ColumnPairMetric
from great_expectations.metrics.metric_name import MetricNameSuffix
from great_expectations.metrics.metric_results import MetricResult


class ColumnPairValuesInSetUnexpectedCountResult(MetricResult[int]): ...


class ColumnPairValuesInSetUnexpectedCount(
    ColumnPairMetric[ColumnPairValuesInSetUnexpectedCountResult]
):
    """Count of values in a column pair that are not in the set of expected values"""

    name = f"column_pair_values.in_set.{MetricNameSuffix.UNEXPECTED_COUNT.value}"
    value_pairs_set: set[tuple[Any, Any]]



from unittest.mock import ANY

import pandas as pd
import pytest

import great_expectations.expectations as gxe
from great_expectations.core.result_format import ResultFormat
from great_expectations.datasource.fluent.interfaces import Batch
from tests.integration.conftest import parameterize_batch_for_data_sources
from tests.integration.data_sources_and_expectations.test_canonical_expectations import (
    JUST_PANDAS_DATA_SOURCES,
    SQL_DATA_SOURCES,
)
from tests.integration.test_utils.data_source_config import (
    PandasDataFrameDatasourceTestConfig,
    PandasFilesystemCsvDatasourceTestConfig,
    SparkFilesystemCsvDatasourceTestConfig,
)

NON_NULL_COLUMN = "none_nulls"
ALL_NULL_COLUMN = "oops_all_nulls"
MOSTLY_NULL_COLUMN = "mostly_nulls"

DATA = pd.DataFrame(
    {
        NON_NULL_COLUMN: [1, 2, 3, 4, 5],
        MOSTLY_NULL_COLUMN: pd.Series([1, None, None, None, None], dtype="object"),
        ALL_NULL_COLUMN: pd.Series([None, None, None, None, None], dtype="object"),
    },
)

try:
    from great_expectations.compatibility.pyspark import types as PYSPARK_TYPES

    SPARK_COLUMN_TYPES = {
        NON_NULL_COLUMN: PYSPARK_TYPES.IntegerType,
        MOSTLY_NULL_COLUMN: PYSPARK_TYPES.IntegerType,
        ALL_NULL_COLUMN: PYSPARK_TYPES.IntegerType,
    }
except ModuleNotFoundError:
    SPARK_COLUMN_TYPES = {}


@parameterize_batch_for_data_sources(
    data_source_configs=[PandasDataFrameDatasourceTestConfig()], data=DATA
)
def test_failure_pandas_dataframe(batch_for_datasource: Batch) -> None:
    expectation = gxe.ExpectColumnValuesToNotBeNull(column=MOSTLY_NULL_COLUMN)
    result = batch_for_datasource.validate(expectation, result_format=ResultFormat.COMPLETE)
    assert not result.success
    assert result.to_json_dict()["result"] == {
        "element_count": 5,
        "unexpected_count": 4,
        "unexpected_percent": 80.0,
        "partial_unexpected_list": [None, None, None, None],
        "partial_unexpected_index_list": [1, 2, 3, 4],
        "partial_unexpected_counts": [
            {"count": 4, "value": None},
        ],
        "unexpected_list": [None, None, None, None],
        "unexpected_index_list": [1, 2, 3, 4],
        "unexpected_index_query": "df.filter(items=[1, 2, 3, 4], axis=0)",
    }


@parameterize_batch_for_data_sources(
    data_source_configs=[PandasFilesystemCsvDatasourceTestConfig()], data=DATA
)
def test_failure_pandas_csv(batch_for_datasource: Batch) -> None:
    expectation = gxe.ExpectColumnValuesToNotBeNull(column=MOSTLY_NULL_COLUMN)
    result = batch_for_datasource.validate(expectation, result_format=ResultFormat.COMPLETE)
    assert not result.success
    assert result.to_json_dict()["result"] == {
        "element_count": 5,
        "unexpected_count": 4,
        "unexpected_percent": 80.0,
        "partial_unexpected_list": [None, None, None, None],
        "partial_unexpected_index_list": [1, 2, 3, 4],
        "partial_unexpected_counts": [
            # TODO: NOT THIS
            {"count": 1, "value": None},
            {"count": 1, "value": None},
            {"count": 1, "value": None},
            {"count": 1, "value": None},
        ],
        "unexpected_list": [None, None, None, None],
        "unexpected_index_list": [1, 2, 3, 4],
        "unexpected_index_query": "df.filter(items=[1, 2, 3, 4], axis=0)",
    }


@parameterize_batch_for_data_sources(
    data_source_configs=[
        SparkFilesystemCsvDatasourceTestConfig(
            column_types=SPARK_COLUMN_TYPES,
        )
    ],
    data=DATA,
)
def test_failure_spark(batch_for_datasource: Batch) -> None:
    expectation = gxe.ExpectColumnValuesToNotBeNull(column=MOSTLY_NULL_COLUMN)
    result = batch_for_datasource.validate(expectation, result_format=ResultFormat.COMPLETE)
    assert not result.success
    assert result.to_json_dict()["result"] == {
        "element_count": 5,
        "unexpected_count": 4,
        "unexpected_percent": 80.0,
        "partial_unexpected_list": [None, None, None, None],
        "partial_unexpected_counts": [
            {"count": 4, "value": None},
        ],
        "unexpected_list": [None, None, None, None],
        "unexpected_index_query": ANY,
    }


@parameterize_batch_for_data_sources(data_source_configs=SQL_DATA_SOURCES, data=DATA)
def test_failure_complete_sql(batch_for_datasource: Batch) -> None:
    expectation = gxe.ExpectColumnValuesToNotBeNull(column=MOSTLY_NULL_COLUMN)
    result = batch_for_datasource.validate(expectation, result_format=ResultFormat.COMPLETE)
    assert not result.success
    assert result.to_json_dict()["result"] == {
        "element_count": 5,
        "unexpected_count": 4,
        "unexpected_percent": 80.0,
        "partial_unexpected_list": [None, None, None, None],
        "partial_unexpected_counts": [
            {"count": 4, "value": None},
        ],
        "unexpected_list": [None, None, None, None],
        "unexpected_index_query": ANY,
    }


@pytest.mark.parametrize(
    "expectation",
    [
        pytest.param(
            gxe.ExpectColumnValuesToNotBeNull(column=NON_NULL_COLUMN),
            id="no_nulls",
        ),
        pytest.param(
            gxe.ExpectColumnValuesToNotBeNull(column=MOSTLY_NULL_COLUMN, mostly=0.2),
            id="mostly",
        ),
    ],
)
@parameterize_batch_for_data_sources(data_source_configs=JUST_PANDAS_DATA_SOURCES, data=DATA)
def test_success(
    batch_for_datasource: Batch,
    expectation: gxe.ExpectColumnValuesToNotBeNull,
) -> None:
    result = batch_for_datasource.validate(expectation)
    assert result.success


@pytest.mark.parametrize(
    "expectation",
    [
        pytest.param(
            gxe.ExpectColumnValuesToNotBeNull(column=ALL_NULL_COLUMN),
            id="no_nulls",
        ),
        pytest.param(
            gxe.ExpectColumnValuesToNotBeNull(column=MOSTLY_NULL_COLUMN, mostly=0.3),
            id="mostly_not_met",
        ),
    ],
)
@parameterize_batch_for_data_sources(data_source_configs=JUST_PANDAS_DATA_SOURCES, data=DATA)
def test_failure(
    batch_for_datasource: Batch,
    expectation: gxe.ExpectColumnValuesToNotBeNull,
) -> None:
    result = batch_for_datasource.validate(expectation)
    assert not result.success



from __future__ import annotations

import copy
import logging
from abc import ABC, abstractmethod
from dataclasses import asdict, dataclass
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Iterable,
    List,
    Optional,
    Set,
    Tuple,
    Union,
)

import great_expectations.exceptions as gx_exceptions
from great_expectations.compatibility.typing_extensions import override
from great_expectations.core.batch_manager import BatchManager
from great_expectations.core.metric_domain_types import MetricDomainTypes
from great_expectations.expectations.registry import get_metric_provider
from great_expectations.expectations.row_conditions import (
    RowCondition,
    RowConditionParserType,
)
from great_expectations.types import DictDot
from great_expectations.util import (
    convert_to_json_serializable,  # noqa: TID251 # FIXME CoP
    filter_properties_dict,
)
from great_expectations.validator.computed_metric import MetricValue  # noqa: TC001 # FIXME CoP

if TYPE_CHECKING:
    from great_expectations.compatibility.pyspark import functions as F
    from great_expectations.compatibility.sqlalchemy import sqlalchemy as sa
    from great_expectations.core.batch import (
        BatchData,
        BatchDataType,
        BatchDataUnion,
        BatchMarkers,
        BatchSpec,
    )
    from great_expectations.expectations.metrics.metric_provider import MetricProvider
    from great_expectations.validator.metric_configuration import (
        MetricConfiguration,
        MetricConfigurationID,
    )
    from great_expectations.validator.validator import Validator

logger = logging.getLogger(__name__)


class NoOpDict:
    def __getitem__(self, item):
        return None

    def __setitem__(self, key, value):
        return None

    # noinspection PyMethodMayBeStatic,PyUnusedLocal
    def update(self, value):
        return None


@dataclass(frozen=True)
class MetricComputationConfiguration(DictDot):
    """
    MetricComputationConfiguration is a "dataclass" object, which holds components required for metric computation.
    """  # noqa: E501 # FIXME CoP

    metric_configuration: MetricConfiguration
    metric_fn: sa.func | F  # type: ignore[valid-type] # FIXME CoP
    metric_provider_kwargs: dict
    compute_domain_kwargs: Optional[dict] = None
    accessor_domain_kwargs: Optional[dict] = None

    @override
    def to_dict(self) -> dict:
        """Returns: this MetricComputationConfiguration as a Python dictionary

        Returns:
            (dict) representation of present object
        """
        return asdict(self)

    def to_json_dict(self) -> dict:
        """Returns: this MetricComputationConfiguration as a JSON dictionary

        Returns:
            (dict) representation of present object as JSON-compatible Python dictionary
        """
        return convert_to_json_serializable(data=self.to_dict())


@dataclass
class PartitionDomainKwargs:
    """compute_domain_kwargs, accessor_domain_kwargs when partitioned from domain_kwargs

    The union of compute_domain_kwargs, accessor_domain_kwargs is the input domain_kwargs
    """

    compute: dict
    accessor: dict


class ExecutionEngine(ABC):
    """ExecutionEngine defines interfaces and provides common methods for loading Batch of data and compute metrics.

    ExecutionEngine is the parent class of every backend-specific computational class, tasked with loading Batch of
    data and computing metrics.  Each subclass (corresponding to Pandas, Spark, SQLAlchemy, and any other computational
    components) utilizes generally-applicable methods defined herein, while implementing required backend-specific
    interfaces.  ExecutionEngine base class also performs the important task of translating cloud storage resource URLs
    to format and protocol compatible with given computational mechanism (e.g, Pandas, Spark).  Then ExecutionEngine
    subclasses access the referenced data according to their paritcular compatible protocol and return Batch of data.

    In order to obtain Batch of data, ExecutionEngine (through implementation of key interface methods by subclasses)
    gets data records and also provides access references so that different aspects of data can be loaded at once.
    ExecutionEngine uses BatchManager for Batch caching in order to reduce load on computational backends.

    Crucially, ExecutionEngine serves as focal point for resolving (i.e., computing) metrics.  Wherever opportunities
    arize to bundle multiple metric computations (e.g., SQLAlchemy, Spark), ExecutionEngine utilizes subclasses in order
    to provide specific functionality (bundling of computation is available only for "deferred execution" computational
    systems, such as SQLAlchemy and Spark; it is not available for Pandas, because Pandas computations are immediate).

    Finally, ExecutionEngine defines interfaces for Batch data sampling and partitioning Batch of data along defined axes.

    Constructor builds an ExecutionEngine, using provided configuration options (instatiation is done by child classes).

    Args:
        name: (str) name of this ExecutionEngine
        caching: (Boolean) if True (default), then resolved (computed) metrics are added to local in-memory cache.
        batch_spec_defaults: dictionary of BatchSpec overrides (useful for amending configuration at runtime).
        batch_data_dict: dictionary of Batch objects with corresponding IDs as keys supplied at initialization time
        validator: Validator object (optional) -- not utilized in V3 and later versions
    """  # noqa: E501 # FIXME CoP

    recognized_batch_spec_defaults: Set[str] = set()

    def __init__(
        self,
        name: Optional[str] = None,
        caching: bool = True,
        batch_spec_defaults: Optional[dict] = None,
        batch_data_dict: Optional[dict] = None,
        validator: Optional[Validator] = None,
    ) -> None:
        self.name = name
        self._validator = validator

        # NOTE: using caching makes the strong assumption that the user will not modify the core data store  # noqa: E501 # FIXME CoP
        # (e.g. self.spark_df) over the lifetime of the dataset instance
        self._caching = caching
        # NOTE: 20200918 - this is a naive cache; update.
        if self._caching:
            self._metric_cache: Union[Dict, NoOpDict] = {}
        else:
            self._metric_cache = NoOpDict()

        if batch_spec_defaults is None:
            batch_spec_defaults = {}

        batch_spec_defaults_keys = set(batch_spec_defaults.keys())
        if not batch_spec_defaults_keys <= self.recognized_batch_spec_defaults:
            logger.warning(
                f"""Unrecognized batch_spec_default(s): \
{batch_spec_defaults_keys - self.recognized_batch_spec_defaults!s}
"""
            )

        self._batch_spec_defaults = {
            key: value
            for key, value in batch_spec_defaults.items()
            if key in self.recognized_batch_spec_defaults
        }

        self._batch_manager = BatchManager(execution_engine=self)

        if batch_data_dict is None:
            batch_data_dict = {}

        self._load_batch_data_from_dict(batch_data_dict=batch_data_dict)

        # Gather the call arguments of the present function (and add the "class_name"), filter out the Falsy values, and  # noqa: E501 # FIXME CoP
        # set the instance "_config" variable equal to the resulting dictionary.
        self._config = {
            "name": name,
            "caching": caching,
            "batch_spec_defaults": batch_spec_defaults,
            "batch_data_dict": batch_data_dict,
            "validator": validator,
            "module_name": self.__class__.__module__,
            "class_name": self.__class__.__name__,
        }
        filter_properties_dict(properties=self._config, clean_falsy=True, inplace=True)

    def configure_validator(  # noqa: B027 # empty-method-without-abstract-decorator
        self, validator
    ) -> None:
        """Optionally configure the validator as appropriate for the execution engine."""
        pass

    @property
    def config(self) -> dict:
        return self._config

    @property
    def dialect(self):
        return None

    @property
    def batch_manager(self) -> BatchManager:
        """Getter for batch_manager"""
        return self._batch_manager

    def _load_batch_data_from_dict(self, batch_data_dict: Dict[str, BatchDataType]) -> None:
        """
        Loads all data in batch_data_dict using cache_batch_data
        """
        batch_id: str
        batch_data: BatchDataType
        for batch_id, batch_data in batch_data_dict.items():
            self.load_batch_data(batch_id=batch_id, batch_data=batch_data)  # type: ignore[arg-type] # FIXME CoP

    def load_batch_data(self, batch_id: str, batch_data: BatchDataUnion) -> None:
        self._batch_manager.save_batch_data(batch_id=batch_id, batch_data=batch_data)

    def get_batch_data(
        self,
        batch_spec: BatchSpec,
    ) -> Any:
        """Interprets batch_data and returns the appropriate data.

        This method is primarily useful for utility cases (e.g. testing) where
        data is being fetched without a DataConnector and metadata like
        batch_markers is unwanted

        Note: this method is currently a thin wrapper for get_batch_data_and_markers.
        It simply suppresses the batch_markers.
        """
        batch_data, _ = self.get_batch_data_and_markers(batch_spec)
        return batch_data

    @abstractmethod
    def get_batch_data_and_markers(self, batch_spec) -> Tuple[BatchData, BatchMarkers]:
        raise NotImplementedError

    def resolve_metrics(
        self,
        metrics_to_resolve: Iterable[MetricConfiguration],
        metrics: Optional[dict[MetricConfigurationID, MetricValue]] = None,
        runtime_configuration: Optional[dict] = None,
    ) -> dict[MetricConfigurationID, MetricValue]:
        """resolve_metrics is the main entrypoint for an execution engine. The execution engine will compute the value
        of the provided metrics.

        Args:
            metrics_to_resolve: the metrics to evaluate
            metrics: already-computed metrics currently available to the engine
            runtime_configuration: runtime configuration information

        Returns:
            resolved_metrics (Dict): a dictionary with the values for the metrics that have just been resolved.
        """  # noqa: E501 # FIXME CoP
        if not metrics_to_resolve:
            return metrics or {}

        metric_fn_direct_configurations: List[MetricComputationConfiguration]
        metric_fn_bundle_configurations: List[MetricComputationConfiguration]
        (
            metric_fn_direct_configurations,
            metric_fn_bundle_configurations,
        ) = self._build_direct_and_bundled_metric_computation_configurations(
            metrics_to_resolve=metrics_to_resolve,
            metrics=metrics,
            runtime_configuration=runtime_configuration,
        )
        return self._process_direct_and_bundled_metric_computation_configurations(
            metric_fn_direct_configurations=metric_fn_direct_configurations,
            metric_fn_bundle_configurations=metric_fn_bundle_configurations,
        )

    def resolve_metric_bundle(self, metric_fn_bundle) -> dict[MetricConfigurationID, MetricValue]:
        """Resolve a bundle of metrics with the same compute Domain as part of a single trip to the compute engine."""  # noqa: E501 # FIXME CoP
        raise NotImplementedError

    def get_domain_records(
        self,
        domain_kwargs: dict,
    ) -> Any:
        """get_domain_records() is an interface method, which computes the full-access data (dataframe or selectable) for computing metrics based on the given domain_kwargs and specific engine semantics.

        Args:
            domain_kwargs (dict) - A dictionary consisting of the Domain kwargs specifying which data to obtain

        Returns:
            data corresponding to the compute domain
        """  # noqa: E501 # FIXME CoP

        raise NotImplementedError

    def get_compute_domain(
        self,
        domain_kwargs: dict,
        domain_type: Union[str, MetricDomainTypes],
        accessor_keys: Optional[Iterable[str]] = None,
    ) -> Tuple[Any, dict, dict]:
        """get_compute_domain() is an interface method, which computes the optimal domain_kwargs for computing metrics based on the given domain_kwargs and specific engine semantics.

        Args:
            domain_kwargs (dict): a dictionary consisting of the Domain kwargs specifying which data to obtain
            domain_type (str or MetricDomainTypes): an Enum value indicating which metric Domain the user would like \
            to be using, or a corresponding string value representing it.  String types include "column", \
            "column_pair", "table", and "other".  Enum types include capitalized versions of these from the class \
            MetricDomainTypes.
            accessor_keys (str iterable): keys that are part of the compute Domain but should be ignored when \
            describing the Domain and simply transferred with their associated values into accessor_domain_kwargs.

        Returns:
            A tuple consisting of three elements:

            1. data corresponding to the compute domain;
            2. a modified copy of domain_kwargs describing the Domain of the data returned in (1);
            3. a dictionary describing the access instructions for data elements included in the compute domain \
                (e.g. specific column name).

            In general, the union of the compute_domain_kwargs and accessor_domain_kwargs will be the same as the
            domain_kwargs provided to this method.
        """  # noqa: E501 # FIXME CoP

        raise NotImplementedError

    def add_column_row_condition(
        self, domain_kwargs, column_name=None, filter_null=True, filter_nan=False
    ):
        """
        Add a row condition for handling null filter.

        Args:
            domain_kwargs: the Domain kwargs to use as the base and to which to add the condition
            column_name: if provided, use this name to add the condition; otherwise, will use "column" key from
                table_domain_kwargs
            filter_null: if true, add a filter for null values
            filter_nan: if true, add a filter for nan values
        """  # noqa: E501 # FIXME CoP
        if filter_null is False and filter_nan is False:
            logger.warning("add_column_row_condition called with no filter condition requested")
            return domain_kwargs

        if filter_nan:
            raise gx_exceptions.GreatExpectationsError(  # noqa: TRY003 # FIXME CoP
                "Base ExecutionEngine does not support adding nan condition filters"
            )

        new_domain_kwargs = copy.deepcopy(domain_kwargs)
        assert "column" in domain_kwargs or column_name is not None, (
            "No column provided: A column must be provided "
            "in domain_kwargs or in the column_name parameter"
        )
        if column_name is not None:
            column = column_name
        else:
            column = domain_kwargs["column"]

        row_condition = RowCondition(
            condition=f'col("{column}").notnull()',
            condition_type=RowConditionParserType.GE,
        )
        new_domain_kwargs.setdefault("filter_conditions", []).append(row_condition)
        return new_domain_kwargs

    def _build_direct_and_bundled_metric_computation_configurations(
        self,
        metrics_to_resolve: Iterable[MetricConfiguration],
        metrics: Optional[dict[MetricConfigurationID, MetricValue]] = None,
        runtime_configuration: Optional[dict] = None,
    ) -> Tuple[
        List[MetricComputationConfiguration],
        List[MetricComputationConfiguration],
    ]:
        """
        This method organizes "metrics_to_resolve" ("MetricConfiguration" objects) into two lists: direct and bundled.
        Directly-computable "MetricConfiguration" must have non-NULL metric function ("metric_fn").  Aggregate metrics
        have NULL metric function, but non-NULL partial metric function ("metric_partial_fn"); aggregates are bundled.

        See documentation in "MetricProvider._register_metric_functions()" for in-depth description of this mechanism.

        Args:
            metrics_to_resolve: the metrics to evaluate
            metrics: already-computed metrics currently available to the engine
            runtime_configuration: runtime configuration information

        Returns:
            Tuple with two elements: directly-computable and bundled "MetricComputationConfiguration" objects
        """  # noqa: E501 # FIXME CoP
        metric_fn_direct_configurations: List[MetricComputationConfiguration] = []
        metric_fn_bundle_configurations: List[MetricComputationConfiguration] = []

        if not metrics_to_resolve:
            return (
                metric_fn_direct_configurations,
                metric_fn_bundle_configurations,
            )

        if metrics is None:
            metrics = {}

        resolved_metric_dependencies_by_metric_name: Dict[
            str, Union[MetricValue, Tuple[Any, dict, dict]]
        ]
        metric_class: MetricProvider
        metric_fn: Union[Callable, None]
        metric_aggregate_fn: sa.func | F  # type: ignore[valid-type] # FIXME CoP
        metric_provider_kwargs: dict
        compute_domain_kwargs: dict
        accessor_domain_kwargs: dict
        metric_to_resolve: MetricConfiguration
        for metric_to_resolve in metrics_to_resolve:
            resolved_metric_dependencies_by_metric_name = (
                self._get_computed_metric_evaluation_dependencies_by_metric_name(
                    metric_to_resolve=metric_to_resolve,
                    metrics=metrics,
                )
            )
            metric_class, metric_fn = get_metric_provider(
                metric_name=metric_to_resolve.metric_name, execution_engine=self
            )
            metric_provider_kwargs = {
                "cls": metric_class,
                "execution_engine": self,
                "metric_domain_kwargs": metric_to_resolve.metric_domain_kwargs,
                "metric_value_kwargs": metric_to_resolve.metric_value_kwargs,
                "metrics": resolved_metric_dependencies_by_metric_name,
                "runtime_configuration": runtime_configuration,
            }
            if metric_fn is None:
                try:
                    (
                        metric_aggregate_fn,
                        compute_domain_kwargs,
                        accessor_domain_kwargs,
                    ) = resolved_metric_dependencies_by_metric_name.pop("metric_partial_fn")
                except KeyError as e:
                    raise gx_exceptions.MetricError(
                        message=f'Missing metric dependency: {e!s} for metric "{metric_to_resolve.metric_name}".'  # noqa: E501 # FIXME CoP
                    )

                metric_fn_bundle_configurations.append(
                    MetricComputationConfiguration(
                        metric_configuration=metric_to_resolve,
                        metric_fn=metric_aggregate_fn,
                        metric_provider_kwargs=metric_provider_kwargs,
                        compute_domain_kwargs=compute_domain_kwargs,
                        accessor_domain_kwargs=accessor_domain_kwargs,
                    )
                )
            else:
                metric_fn_direct_configurations.append(
                    MetricComputationConfiguration(
                        metric_configuration=metric_to_resolve,
                        metric_fn=metric_fn,
                        metric_provider_kwargs=metric_provider_kwargs,
                    )
                )

        return (
            metric_fn_direct_configurations,
            metric_fn_bundle_configurations,
        )

    def _get_computed_metric_evaluation_dependencies_by_metric_name(
        self,
        metric_to_resolve: MetricConfiguration,
        metrics: Dict[MetricConfigurationID, MetricValue],
    ) -> Dict[str, Union[MetricValue, Tuple[Any, dict, dict]]]:
        """
        Gathers resolved (already computed) evaluation dependencies of metric-to-resolve (not yet computed)
        "MetricConfiguration" object by "metric_name" property of resolved "MetricConfiguration" objects.

        Args:
            metric_to_resolve: dependent (not yet resolved) "MetricConfiguration" object
            metrics: resolved (already computed) "MetricConfiguration" objects keyd by ID of that object

        Returns:
            Dictionary keyed by "metric_name" with values as computed metric or partial bundling information tuple
        """  # noqa: E501 # FIXME CoP
        metric_dependencies_by_metric_name: Dict[
            str, Union[MetricValue, Tuple[Any, dict, dict]]
        ] = {}

        metric_name: str
        metric_configuration: MetricConfiguration
        for (
            metric_name,
            metric_configuration,
        ) in metric_to_resolve.metric_dependencies.items():
            if metric_configuration.id in metrics:
                metric_dependencies_by_metric_name[metric_name] = metrics[metric_configuration.id]
            elif self._caching and metric_configuration.id in self._metric_cache:  # type: ignore[operator] # TODO: update NoOpDict
                metric_dependencies_by_metric_name[metric_name] = self._metric_cache[
                    metric_configuration.id
                ]
            else:
                raise gx_exceptions.MetricError(
                    message=f'Missing metric dependency: "{metric_name}" for metric "{metric_to_resolve.metric_name}".'  # noqa: E501 # FIXME CoP
                )

        return metric_dependencies_by_metric_name

    def _process_direct_and_bundled_metric_computation_configurations(
        self,
        metric_fn_direct_configurations: List[MetricComputationConfiguration],
        metric_fn_bundle_configurations: List[MetricComputationConfiguration],
    ) -> dict[MetricConfigurationID, MetricValue]:
        """
        This method processes directly-computable and bundled "MetricComputationConfiguration" objects.

        Args:
            metric_fn_direct_configurations: directly-computable "MetricComputationConfiguration" objects
            metric_fn_bundle_configurations: bundled "MetricComputationConfiguration" objects (column aggregates)

        Returns:
            resolved_metrics (Dict): a dictionary with the values for the metrics that have just been resolved.
        """  # noqa: E501 # FIXME CoP
        resolved_metrics: dict[MetricConfigurationID, MetricValue] = {}

        metric_computation_configuration: MetricComputationConfiguration

        for metric_computation_configuration in metric_fn_direct_configurations:
            try:
                resolved_metrics[metric_computation_configuration.metric_configuration.id] = (
                    metric_computation_configuration.metric_fn(  # type: ignore[misc] # F not callable
                        **metric_computation_configuration.metric_provider_kwargs
                    )
                )
            except Exception as e:
                raise gx_exceptions.MetricResolutionError(
                    message=str(e),
                    failed_metrics=(metric_computation_configuration.metric_configuration,),
                ) from e

        try:
            # an engine-specific way of computing metrics together
            resolved_metric_bundle: dict[MetricConfigurationID, MetricValue] = (
                self.resolve_metric_bundle(metric_fn_bundle=metric_fn_bundle_configurations)
            )
            resolved_metrics.update(resolved_metric_bundle)
        except Exception as e:
            raise gx_exceptions.MetricResolutionError(
                message=str(e),
                failed_metrics=[
                    metric_computation_configuration.metric_configuration
                    for metric_computation_configuration in metric_fn_bundle_configurations
                ],
            ) from e

        if self._caching:
            self._metric_cache.update(resolved_metrics)

        return resolved_metrics

    def _partition_domain_kwargs(
        self,
        domain_kwargs: Dict[str, Any],
        domain_type: Union[str, MetricDomainTypes],
        accessor_keys: Optional[Iterable[str]] = None,
    ) -> PartitionDomainKwargs:
        """Partition domain_kwargs for all Domain types into compute and accessor Domain kwargs.

        Args:
            domain_kwargs: A dictionary consisting of the Domain kwargs specifying which data to obtain
            domain_type: an Enum value indicating which metric Domain the user would
            like to be using, or a corresponding string value representing it. String types include "identity",
            "column", "column_pair", "table" and "other". Enum types include capitalized versions of these from the
            class MetricDomainTypes.
            accessor_keys: keys that are part of the compute Domain but should be ignored when
            describing the Domain and simply transferred with their associated values into accessor_domain_kwargs.

        Returns:
            compute_domain_kwargs, accessor_domain_kwargs from domain_kwargs
            The union of compute_domain_kwargs, accessor_domain_kwargs is the input domain_kwargs
        """  # noqa: E501 # FIXME CoP
        # Extracting value from enum if it is given for future computation
        domain_type = MetricDomainTypes(domain_type)

        # Warning user if accessor keys are in any Domain that is not of type table, will be ignored
        if (
            domain_type != MetricDomainTypes.TABLE
            and accessor_keys is not None
            and len(list(accessor_keys)) > 0
        ):
            logger.warning('Accessor keys ignored since Metric Domain Type is not "table"')

        partition_domain_kwargs: PartitionDomainKwargs
        if domain_type == MetricDomainTypes.TABLE:
            partition_domain_kwargs = self._partition_table_metric_domain_kwargs(
                domain_kwargs, domain_type, accessor_keys
            )

        elif domain_type == MetricDomainTypes.COLUMN:
            partition_domain_kwargs = self._partition_column_metric_domain_kwargs(
                domain_kwargs,
                domain_type,
            )

        elif domain_type == MetricDomainTypes.COLUMN_PAIR:
            partition_domain_kwargs = self._partition_column_pair_metric_domain_kwargs(
                domain_kwargs,
                domain_type,
            )

        elif domain_type == MetricDomainTypes.MULTICOLUMN:
            partition_domain_kwargs = self._partition_multi_column_metric_domain_kwargs(
                domain_kwargs,
                domain_type,
            )
        else:
            compute_domain_kwargs = copy.deepcopy(domain_kwargs)
            accessor_domain_kwargs: Dict[str, Any] = {}
            partition_domain_kwargs = PartitionDomainKwargs(
                compute_domain_kwargs, accessor_domain_kwargs
            )

        return partition_domain_kwargs

    @staticmethod
    def _partition_table_metric_domain_kwargs(
        domain_kwargs: dict,
        domain_type: MetricDomainTypes,
        accessor_keys: Optional[Iterable[str]] = None,
    ) -> PartitionDomainKwargs:
        """Partition domain_kwargs for table Domain types into compute and accessor Domain kwargs.

        Args:
            domain_kwargs: A dictionary consisting of the Domain kwargs specifying which data to obtain
            domain_type: an Enum value indicating which metric Domain the user would
            like to be using.
            accessor_keys: keys that are part of the compute Domain but should be ignored when
            describing the Domain and simply transferred with their associated values into accessor_domain_kwargs.

        Returns:
            compute_domain_kwargs, accessor_domain_kwargs from domain_kwargs
            The union of compute_domain_kwargs, accessor_domain_kwargs is the input domain_kwargs
        """  # noqa: E501 # FIXME CoP
        assert domain_type == MetricDomainTypes.TABLE, (
            "This method only supports MetricDomainTypes.TABLE"
        )

        compute_domain_kwargs: Dict = copy.deepcopy(domain_kwargs)
        accessor_domain_kwargs: Dict = {}

        if accessor_keys is not None and len(list(accessor_keys)) > 0:
            for key in accessor_keys:
                accessor_domain_kwargs[key] = compute_domain_kwargs.pop(key)
        if len(domain_kwargs.keys()) > 0:
            # Warn user if kwarg not "normal".
            unexpected_keys: set = set(compute_domain_kwargs.keys()).difference(
                {
                    "batch_id",
                    "table",
                    "row_condition",
                    "condition_parser",
                }
            )
            if len(unexpected_keys) > 0:
                unexpected_keys_str: str = ", ".join(
                    map(lambda element: f'"{element}"', unexpected_keys)
                )
                logger.warning(
                    f"""Unexpected key(s) {unexpected_keys_str} found in domain_kwargs for Domain type "{domain_type.value}"."""  # noqa: E501 # FIXME CoP
                )

        return PartitionDomainKwargs(compute_domain_kwargs, accessor_domain_kwargs)

    @staticmethod
    def _partition_column_metric_domain_kwargs(
        domain_kwargs: dict,
        domain_type: MetricDomainTypes,
    ) -> PartitionDomainKwargs:
        """Partition domain_kwargs for column Domain types into compute and accessor Domain kwargs.

        Args:
            domain_kwargs: A dictionary consisting of the Domain kwargs specifying which data to obtain
            domain_type: an Enum value indicating which metric Domain the user would
            like to be using.

        Returns:
            compute_domain_kwargs, accessor_domain_kwargs from domain_kwargs
            The union of compute_domain_kwargs, accessor_domain_kwargs is the input domain_kwargs
        """  # noqa: E501 # FIXME CoP
        assert domain_type == MetricDomainTypes.COLUMN, (
            "This method only supports MetricDomainTypes.COLUMN"
        )

        compute_domain_kwargs: Dict = copy.deepcopy(domain_kwargs)
        accessor_domain_kwargs: Dict = {}

        if "column" not in compute_domain_kwargs:
            raise gx_exceptions.GreatExpectationsError(  # noqa: TRY003 # FIXME CoP
                "Column not provided in compute_domain_kwargs"
            )

        accessor_domain_kwargs["column"] = compute_domain_kwargs.pop("column")

        return PartitionDomainKwargs(compute_domain_kwargs, accessor_domain_kwargs)

    @staticmethod
    def _partition_column_pair_metric_domain_kwargs(
        domain_kwargs: dict,
        domain_type: MetricDomainTypes,
    ) -> PartitionDomainKwargs:
        """Partition domain_kwargs for column pair Domain types into compute and accessor Domain kwargs.

        Args:
            domain_kwargs: A dictionary consisting of the Domain kwargs specifying which data to obtain
            domain_type: an Enum value indicating which metric Domain the user would
            like to be using.

        Returns:
            compute_domain_kwargs, accessor_domain_kwargs from domain_kwargs
            The union of compute_domain_kwargs, accessor_domain_kwargs is the input domain_kwargs
        """  # noqa: E501 # FIXME CoP
        assert domain_type == MetricDomainTypes.COLUMN_PAIR, (
            "This method only supports MetricDomainTypes.COLUMN_PAIR"
        )

        compute_domain_kwargs: Dict = copy.deepcopy(domain_kwargs)
        accessor_domain_kwargs: Dict = {}

        if not ("column_A" in domain_kwargs and "column_B" in domain_kwargs):
            raise gx_exceptions.GreatExpectationsError(  # noqa: TRY003 # FIXME CoP
                "column_A or column_B not found within domain_kwargs"
            )

        accessor_domain_kwargs["column_A"] = compute_domain_kwargs.pop("column_A")
        accessor_domain_kwargs["column_B"] = compute_domain_kwargs.pop("column_B")

        return PartitionDomainKwargs(compute_domain_kwargs, accessor_domain_kwargs)

    @staticmethod
    def _partition_multi_column_metric_domain_kwargs(
        domain_kwargs: dict,
        domain_type: MetricDomainTypes,
    ) -> PartitionDomainKwargs:
        """Partition domain_kwargs for multicolumn Domain types into compute and accessor Domain kwargs.

        Args:
            domain_kwargs: A dictionary consisting of the Domain kwargs specifying which data to obtain
            domain_type: an Enum value indicating which metric Domain the user would
            like to be using.

        Returns:
            compute_domain_kwargs, accessor_domain_kwargs from domain_kwargs
            The union of compute_domain_kwargs, accessor_domain_kwargs is the input domain_kwargs
        """  # noqa: E501 # FIXME CoP
        assert domain_type == MetricDomainTypes.MULTICOLUMN, (
            "This method only supports MetricDomainTypes.MULTICOLUMN"
        )

        compute_domain_kwargs: Dict = copy.deepcopy(domain_kwargs)
        accessor_domain_kwargs: Dict = {}

        if "column_list" not in domain_kwargs:
            raise gx_exceptions.GreatExpectationsError("column_list not found within domain_kwargs")  # noqa: TRY003 # FIXME CoP

        column_list = compute_domain_kwargs.pop("column_list")

        if len(column_list) < 2:  # noqa: PLR2004 # FIXME CoP
            raise gx_exceptions.GreatExpectationsError(  # noqa: TRY003 # FIXME CoP
                "column_list must contain at least 2 columns"
            )

        accessor_domain_kwargs["column_list"] = column_list

        return PartitionDomainKwargs(compute_domain_kwargs, accessor_domain_kwargs)



from typing import List

from tests.integration.integration_test_fixture import IntegrationTestFixture

trino_integration_tests = []

connecting_to_your_data: List[IntegrationTestFixture] = []


trino_integration_tests += connecting_to_your_data



# This file is intended for tests whose functionality is a workaround for deficiencies in upstream libraries.  # noqa: E501 # FIXME CoP
# In an ideal world, it is empty. As fixes come to be, we can replace its items.



from great_expectations.metrics.batch import BatchMetric
from great_expectations.metrics.metric_results import MetricResult


class BatchRowCountResult(MetricResult[int]): ...


class BatchRowCount(BatchMetric[BatchRowCountResult]):
    """Count of rows in a table"""

    name = "table.row_count"
